import heapq
import threading
import time
from functools import total_ordering
from threading import Condition
from typing import Optional, Dict, List, Tuple

from messageflux.iodevices.base import (Message,
                                        InputDeviceManager,
                                        OutputDeviceManager,
                                        OutputDevice,
                                        InputDevice,
                                        InputTransaction,
                                        ReadResult)
from messageflux.iodevices.base.common import MessageBundle
from messageflux.iodevices.base.input_transaction import NULLTransaction

MESSAGE_TIMESTAMP_HEADER = 'message_timestamp'


@total_ordering
class _QueueMessage:
    """
    an object containing a message in the queue
    """

    def __init__(self, message: Message, timestamp: Optional[float] = None):
        self.message = message.copy()
        self.timestamp = timestamp or time.time()
        self._monotonic_time = time.perf_counter_ns()

    def __eq__(self, other):
        if not isinstance(other, _QueueMessage):
            return False
        return self._monotonic_time == other._monotonic_time

    def __lt__(self, other):
        assert isinstance(other, _QueueMessage)
        return self._monotonic_time < other._monotonic_time


class InMemoryInputDevice(InputDevice['InMemoryDeviceManager']):
    """
    an input device which is stored in memory
    """

    class InMemoryTransaction(InputTransaction):
        """
        a transaction object for the in memory device
        """

        def __init__(self, device: 'InMemoryInputDevice', message: _QueueMessage):
            super().__init__(device)
            self._message = message
            self._device: InMemoryInputDevice = device

        def _commit(self):
            pass

        def _rollback(self):
            """
            returns the message back to queue
            """
            self._device._push_to_queue(self._message)

    def __init__(self, manager: 'InMemoryDeviceManager',
                 name: str,
                 queue: List[_QueueMessage],
                 queue_not_empty_condition: Condition):
        super().__init__(manager, name)
        self._queue = queue
        self._queue_not_empty = queue_not_empty_condition

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional[ReadResult]:

        with self._queue_not_empty:
            if self._queue_not_empty.wait_for(lambda: any(self._queue), timeout):
                queue_message = heapq.heappop(self._queue)
                transaction: InputTransaction
                if with_transaction:
                    transaction = self.InMemoryTransaction(self, queue_message)
                else:
                    transaction = NULLTransaction(self)
                device_headers = {MESSAGE_TIMESTAMP_HEADER: queue_message.timestamp}
                return ReadResult(message=queue_message.message.copy(),
                                  device_headers=device_headers,
                                  transaction=transaction)
            else:
                return None

    def _push_to_queue(self, message: _QueueMessage):
        with self._queue_not_empty:
            heapq.heappush(self._queue, message)
            self._queue_not_empty.notify()


class InMemoryOutputDevice(OutputDevice['InMemoryDeviceManager']):
    """
    an output device which is stored in memory
    """

    def __init__(self, manager: 'InMemoryDeviceManager',
                 name: str,
                 queue: List[_QueueMessage],
                 queue_not_empty_condition: Condition):
        super().__init__(manager, name)
        self._queue = queue
        self._queue_not_empty = queue_not_empty_condition

    def _send_message(self, message_bundle: MessageBundle):
        """
        sends a message to the device.

        :param message_bundle: the message bundle to send
        """
        with self._queue_not_empty:
            heapq.heappush(self._queue, _QueueMessage(message_bundle.message))
            self._queue_not_empty.notify()


class InMemoryDeviceManager(InputDeviceManager[InMemoryInputDevice], OutputDeviceManager[InMemoryOutputDevice]):
    """
    the in memory device manager. it serves as both input and output device manager.
    notice that the messages are shared only within the same manager!
    """

    def __init__(self):
        self._queues: Dict[str, Tuple[List[_QueueMessage], Condition]] = {}

    def _get_queue_tuple(self, name: str) -> Tuple[List[_QueueMessage], Condition]:
        res = self._queues.get(name, None)
        if res is not None:
            queue, condition = res
        else:
            queue = []
            condition = Condition()
            self._queues[name] = (queue, condition)

        return queue, condition

    def get_input_device(self, name: str) -> InMemoryInputDevice:
        """
        creates an input device.

        :param name: the name of the input device to create
        :return: the created input device
        """
        queue, condition = self._get_queue_tuple(name)

        return InMemoryInputDevice(self, name, queue, condition)

    def get_output_device(self, name: str) -> InMemoryOutputDevice:
        """
        creates an output device.

        :param name: the name of the output device to create
        :return: the created output device
        """
        queue, condition = self._get_queue_tuple(name)
        return InMemoryOutputDevice(self, name, queue, condition)



import logging
import threading
from abc import ABCMeta, abstractmethod
from time import perf_counter
from typing import Optional, List, TypeVar, Generic

from messageflux.iodevices.base.common import MessageBundle, Message, DeviceHeaders
from messageflux.iodevices.base.input_transaction import InputTransaction, NULLTransaction
from messageflux.utils import StatefulListIterator, AggregatedException

TManagerType = TypeVar('TManagerType', bound='InputDeviceManager')
TInputDeviceType = TypeVar('TInputDeviceType', bound='InputDevice')


class InputDeviceException(AggregatedException):
    """
    a base exception class for all input device related exceptions
    """
    pass


class InputDevice(Generic[TManagerType], metaclass=ABCMeta):
    """
    this is the base class for input devices
    """
    INPUT_DEVICE_NAME_HEADER = "__INPUT_DEVICE_NAME__"

    def __init__(self, manager: TManagerType, name: str):
        """
        :param manager: the input device manager that created this device
        :param name: the name of this device
        """
        self._manager = manager
        self._name = name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def name(self) -> str:
        """
        :return: the name of this device
        """
        return self._name

    @property
    def manager(self) -> TManagerType:
        """
        :return: the input device manager that created this device
        """
        return self._manager

    def read_message(self,
                     cancellation_token: threading.Event,
                     timeout: Optional[float] = None,
                     with_transaction: bool = True) -> Optional['ReadResult']:
        """
        this method returns a message from the device. and makes sure that the input device name header is present

        :param cancellation_token: the cancellation token for this service. this can be used to know if cancellation
        was requested

        :param timeout: an optional timeout (in seconds) to wait for the device to return a message.
        after 'timeout' seconds, if the device doesn't have a message to return, it will return None
        :param with_transaction: 'True' if the device should read message within transaction,
        or 'False' if the message is automatically committed

        :return: a ReadResult object or None if no message was available.
        the device headers can contain extra information about the device that returned the message
        """
        read_result = self._read_message(cancellation_token=cancellation_token,
                                         timeout=timeout,
                                         with_transaction=with_transaction)
        if read_result is not None:
            read_result.device_headers.setdefault(self.INPUT_DEVICE_NAME_HEADER, self.name)

        return read_result

    @abstractmethod
    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional['ReadResult']:
        """
        this method returns a message from the device (should be implemented by child classes)

        :param cancellation_token: the cancellation token for this service. this can be used to know if cancellation
        was requested

        :param timeout: an optional timeout (in seconds) to wait for the device to return a message.
        after 'timeout' seconds, if the device doesn't have a message to return, it will return None

        :param with_transaction: 'True' if the device should read message within transaction,
        or 'False' if the message is automatically committed

        :return: a ReadResult object or None if no message was available.
        the device headers, can contain extra information about the device that returned the message
        """
        pass

    def close(self):
        """
        and optional method that cleans device resources if necessary
        """
        pass


class AggregatedInputDevice(InputDevice[TManagerType]):
    """
    this class is a round-robin input device, that reads from several underlying input devices in order
    """
    # the amount in seconds to sleep between iterations. otherwise, we perform busy wait if all devices are empty
    _SLEEP_BETWEEN_ITERATIONS: float = 0.1

    def __init__(self, manager: TManagerType, inner_devices: List[InputDevice]):
        """
        :param manager: the input device manager that created this device
        :param inner_devices: the list of input devices to read from
        """
        super().__init__(manager=manager, name="AggregateInputDevice")
        self._inner_devices_iterator: StatefulListIterator[InputDevice] = StatefulListIterator(inner_devices)
        self._last_read_device: Optional[InputDevice] = None
        self._logger = logging.getLogger(__name__)

    @property
    def last_read_device(self) -> Optional[InputDevice]:
        """
        :return: the last device that was read (in case it returned data/raised exception)
        """
        return self._last_read_device

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional['ReadResult']:
        """
        this method returns a message from the first device available in inner devices

        :param cancellation_token: the cancellation token for this service. this can be used to know if cancellation
        was requested

        :param timeout: an optional timeout (in seconds) to wait for the device to return a message.
        after 'timeout' seconds, if the device doesn't have a message to return, it will return None

        :param with_transaction: 'True' if the device should read message within transaction,
        or 'False' if the message is automatically committed

        :return: a ReadResult object or None if no message was available.
        the device headers, can contain extra information about the device that returned the message
        """

        end_time = 0.0
        if timeout is not None:
            end_time = perf_counter() + timeout

        while True:
            for inner_device in self._inner_devices_iterator:
                self._last_read_device = inner_device
                read_result = inner_device.read_message(cancellation_token=cancellation_token,
                                                        timeout=0,
                                                        with_transaction=with_transaction)
                if read_result is not None:
                    return read_result

                if timeout is not None and perf_counter() >= end_time:
                    break

            if timeout is not None and perf_counter() >= end_time:
                self._last_read_device = None
                return None
            else:
                # if all the devices were empty, wait before performing another iteration.
                cancellation_token.wait(self._SLEEP_BETWEEN_ITERATIONS)

    def close(self):
        """
        tries to close underlying devices
        """
        for inner_device in self._inner_devices_iterator:
            try:
                inner_device.close()
            except Exception:
                self._logger.exception("Error closing underlying device")


class InputDeviceManager(Generic[TInputDeviceType], metaclass=ABCMeta):
    """
    this is the base class for input device managers. this class is used to create input devices.
    """

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        """
        connects to the device manager
        """
        pass

    def disconnect(self):
        """
        disconnects from the device manager
        """
        pass

    @abstractmethod
    def get_input_device(self, name: str) -> TInputDeviceType:
        """
        creates an input device. should be implemented in child classes

        :param name: the name of the input device to create
        :return: the created input device
        """
        pass

    def get_aggregate_device(self, names: List[str]) -> AggregatedInputDevice:
        """
        creates an aggregated input device on all the devices with given names

        :param names: the names of the devices to create and aggregate
        :return: the AggregatedInputDevice
        """
        inner_devices: List[InputDevice] = []
        for name in names:
            inner_devices.append(self.get_input_device(name))

        return AggregatedInputDevice(manager=self, inner_devices=inner_devices)


class _NullInputDeviceManager(InputDeviceManager):
    """
    this is a stub used to create null device
    """

    def get_input_device(self, name: str) -> InputDevice:
        """
        returns a null device
        """
        return _NullDevice()


class _NullDevice(InputDevice):
    """
    this is a stub for creating a null transaction
    """

    def __init__(self):
        super(_NullDevice, self).__init__(_NullInputDeviceManager(), '__NULL__')

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional['ReadResult']:
        """
        always returns None immediately
        """
        return None


NULL_TRANSACTION = NULLTransaction(_NullDevice())  # this is a const NullTransaction, for convenience


class ReadResult(MessageBundle):
    """
    this class holds the result for "read_message". adds the transaction to the message bundle
    """
    __slots__ = "_transaction"

    def __init__(self,
                 message: Message,
                 device_headers: Optional[DeviceHeaders] = None,
                 transaction: InputTransaction = NULL_TRANSACTION):
        """
        :param message: The Message.
        :param device_headers: Additional Headers that may return data from device, or affect its operation.
        :param transaction: the transaction returned by the reading device
        """
        super(ReadResult, self).__init__(message=message,
                                         device_headers=device_headers)

        self._transaction = transaction

    @property
    def transaction(self) -> InputTransaction:
        """
        the transaction returned by the reading device
        """
        return self._transaction

    def commit(self) -> None:
        """
        commits this read result
        """
        self.transaction.commit()

    def rollback(self) -> None:
        """
        rolls back this read result
        """
        self.transaction.rollback()



import logging
import socket
import ssl
from random import shuffle
from typing import TYPE_CHECKING, List, Optional, Dict, Any, Union, Type

import sys

from messageflux.utils import ThreadLocalMember, KwargsException

DEFAULT_CLIENT_ARGS = {'hostname': socket.gethostname()}
CONNECTION_TYPE_ARG_NAME = '__CONNECTION_TYPE__'

pika_logger = logging.getLogger('pika')

pika_logger.propagate = False
pika_logger.setLevel(logging.WARNING)

for handler in pika_logger.handlers:
    pika_logger.removeHandler(handler)
pika_logger.addHandler(logging.StreamHandler(sys.stdout))

if TYPE_CHECKING:
    from pika import BlockingConnection, PlainCredentials
    from pika.adapters.blocking_connection import BlockingChannel
    from pika.frame import Method as PikaMethod


class RabbitMQDeviceManagerBase:
    """
    base class for rabbitmq device managers

    Notice that pika is imported inside the methods here,
    since it causes trouble when using this device in multiprocess.
    """
    _connection: Union[ThreadLocalMember[Optional['BlockingConnection']],
                       Optional['BlockingConnection']] = ThreadLocalMember(init_value=None)
    _maintenance_channel: Union[ThreadLocalMember[Optional['BlockingChannel']],
                                Optional['BlockingChannel']] = ThreadLocalMember(init_value=None)

    def __init__(self,
                 hosts: Union[List[str], str],
                 user: str,
                 password: str,
                 port: Optional[int] = None,
                 ssl_context: Optional[ssl.SSLContext] = None,
                 virtual_host: Optional[str] = None,
                 client_args: Optional[Dict[str, str]] = None,
                 connection_type: str = "None",
                 heartbeat: int = 300,
                 connection_attempts: int = 5,
                 blocked_connection_timeout: Optional[float] = None):
        """
        This manager used to create RabbitMQ devices (direct queues)

        :param hosts: the hostname or a list of hostnames of the manager
        :param user: the username for the rabbitMQ manager
        :param password: the password for the rabbitMQ manager
        :param port: the port to connect the hosts to
        :param ssl_context: the ssl context to use. None means don't use ssl at all
        :param virtual_host: the virtual host to connect to
        :param client_args: the arguments to create the client with
        :param heartbeat: heartbeat interval for the connection (between 0 and 65536
        :param connection_attempts: Maximum number of retry attempts
        :param blocked_connection_timeout: If not None,
            the value is a non-negative timeout, in seconds, for the
            connection to remain blocked (triggered by Connection.Blocked from
            broker); if the timeout expires before connection becomes unblocked,
            the connection will be torn down, triggering the adapter-specific
            mechanism for informing client app about the closed connection:
            passing `ConnectionBlockedTimeout` exception to on_close_callback
            in asynchronous adapters or raising it in `BlockingConnection`.
        """
        try:
            import pika
        except ImportError as ex:
            raise ImportError('Please Install the required extra: messageflux[rabbitmq]') from ex
        if isinstance(hosts, str):
            hosts = [hosts]
        self._hosts = hosts
        self._user = user
        self._password = password
        self._ssl_context = ssl_context
        if port is None:
            if self._ssl_context is not None:
                port = pika.ConnectionParameters.DEFAULT_SSL_PORT
            else:
                port = pika.ConnectionParameters.DEFAULT_PORT
        self._port = port
        if virtual_host is None:
            virtual_host = pika.ConnectionParameters.DEFAULT_VIRTUAL_HOST
        self._virtual_host = virtual_host
        self._logger = logging.getLogger(__name__)

        self._client_args = (client_args or DEFAULT_CLIENT_ARGS).copy()
        self._client_args[CONNECTION_TYPE_ARG_NAME] = connection_type
        self._client_args["product"] = f"{socket.gethostname()}(Pika)"

        self._heartbeat = max(0, min(heartbeat, 65535))
        self._connection_attempts = connection_attempts
        self._blocked_connection_timeout = blocked_connection_timeout

    def _create_connection(self,
                           credentials: Optional['PlainCredentials'] = None,
                           heartbeat: Optional[int] = None,
                           connection_attempts: Optional[int] = None,
                           client_properties: Optional[Dict[str, str]] = None,
                           virtual_host: Optional[str] = None,
                           blocked_connection_timeout: Optional[float] = None) -> 'BlockingConnection':
        """
        creates a BlockingConnection object

        :param credentials: possible alternate credentials
        :param heartbeat: heartbeat interval for the connection (between 0 and 65536
        :param connection_attempts: Maximum number of retry attempts
        :param client_properties: the client properties for the connection
        :param virtual_host: the virtual host to connect to
        :param blocked_connection_timeout: If not None,
            the value is a non-negative timeout, in seconds, for the
            connection to remain blocked (triggered by Connection.Blocked from
            broker); if the timeout expires before connection becomes unblocked,
            the connection will be torn down, triggering the adapter-specific
            mechanism for informing client app about the closed connection:
            passing `ConnectionBlockedTimeout` exception to on_close_callback
            in asynchronous adapters or raising it in `BlockingConnection`.

        :return: a newly created connection
        """
        try:
            import pika
            from pika.exceptions import AMQPConnectionError
        except ImportError as ex:
            raise ImportError('Please Install the required extra: messageflux[rabbitmq]') from ex

        class SelfClosingBlockingConnection(pika.BlockingConnection):

            @property
            def is_open(self):
                """
                Returns a boolean reporting the current connection state.
                """
                try:
                    self.process_data_events()
                except Exception:
                    pass

                return super().is_open

            def __del__(self):
                """
                Tries to close the connection when it goes out of scope (other thread for example)
                """
                try:
                    if self.is_open:
                        self.close()
                except Exception:
                    pass

        shuffle(self._hosts)
        exceptions = []
        for host in self._hosts:
            try:
                ssl_options: Union[pika.SSLOptions, Type[pika.ConnectionParameters._DEFAULT]]
                if self._ssl_context is not None:
                    ssl_options = pika.SSLOptions(self._ssl_context, host)
                else:
                    ssl_options = pika.ConnectionParameters._DEFAULT

                return SelfClosingBlockingConnection(pika.ConnectionParameters(
                    host=host,
                    port=self._port,
                    ssl_options=ssl_options,
                    credentials=credentials or pika.PlainCredentials(username=self._user, password=self._password),

                    heartbeat=heartbeat or self._heartbeat,
                    connection_attempts=connection_attempts or self._connection_attempts,
                    client_properties=client_properties or self._client_args,
                    virtual_host=virtual_host or self._virtual_host,
                    blocked_connection_timeout=(blocked_connection_timeout or self._blocked_connection_timeout)))

            except AMQPConnectionError as ex:
                message = f"RabbitMQ on host {host} is not available"
                io_device_exception = KwargsException(message)
                if len(self._hosts) == 1:
                    self._logger.exception(message)
                    raise io_device_exception from ex
                else:
                    io_device_exception.__cause__ = ex  # this adds ex as the cause, to use in the aggregate exception
                    exceptions.append(io_device_exception)
                    self._logger.warning(message, exc_info=True)

        raise KwargsException("Couldn't connect to any of the hosts for RabbitMQ.",
                              inner_exceptions=exceptions)

    @property
    def connection(self) -> 'BlockingConnection':
        """
        returns an open and ready connection
        """
        if self._connection is None or not self._connection.is_open:
            self._connection = self._create_connection()

        return self._connection

    def _connect(self):
        if self._connection is None or not self._connection.is_open:
            self._connection = self._create_connection()

    def _disconnect(self):
        if self._connection is not None:
            try:
                self._connection.close()
            except Exception:
                self._logger.warning('Error Closing Device', exc_info=True)

    @property
    def maintenance_channel(self) -> 'BlockingChannel':
        """
        returns a channel used for maintenance (create/bind etc...)
        """
        if (self._maintenance_channel is None or
                not self._maintenance_channel.is_open or not self._maintenance_channel.connection.is_open):
            self._maintenance_channel = self.connection.channel()

        assert self._maintenance_channel is not None
        return self._maintenance_channel

    def create_queue(self,
                     queue_name: str,
                     passive: bool = False,
                     durable: bool = True,
                     exclusive: bool = False,
                     auto_delete: bool = False,
                     arguments: Optional[Dict[str, Any]] = None,
                     direct_bind_to_exchange: Optional[str] = None) -> 'PikaMethod':
        """
        Declare queue, create if needed. This method creates or checks a
        queue. When creating a new queue the client can specify various
        properties that control the durability of the queue and its contents,
        and the level of sharing for the queue.

        Leave the queue name empty for a auto-named queue in RabbitMQ

        :param queue_name: The queue name
        :param passive: Only check to see if the queue exists
        :param durable: Survive reboots of the broker
        :param exclusive: Only allow access by the current connection
        :param auto_delete: Delete after consumer cancels or disconnects
        :param arguments: Custom key/value arguments for the queue
        :param direct_bind_to_exchange: if not None - a 'direct' exchange name to bind to with the queue name
        :returns: Method frame from the Queue.Declare-ok response
        """
        res = self.maintenance_channel.queue_declare(queue=queue_name,
                                                     passive=passive,
                                                     durable=durable,
                                                     exclusive=exclusive,
                                                     auto_delete=auto_delete,
                                                     arguments=arguments)
        if direct_bind_to_exchange is not None:
            self.bind_queue(queue_name, direct_bind_to_exchange, queue_name)
        return res

    def delete_queue(self, queue_name: str, only_if_empty: bool = True) -> 'PikaMethod':
        """
        Deletes a queue from the mq broker.

        :param queue_name: The queue name
        :param only_if_empty: if True will delete the queue only if it is empty (no messages inside)
        :returns: Method frame from the Queue.Declare-ok response
        """
        return self.maintenance_channel.queue_delete(queue_name, if_empty=only_if_empty)

    def get_queue_message_count(self, queue_name: str) -> int:
        """
        returns the current number of messages in queue

        :param queue_name: the queue name to check
        :return: current number of messages in queue
        """
        return self.create_queue(queue_name, passive=True).method.message_count

    def bind_queue(self,
                   queue_name: str,
                   exchange: str,
                   routing_key: Optional[str] = None,
                   arguments: Optional[Dict[str, Any]] = None) -> 'PikaMethod':
        """
        Bind the queue to the specified exchange

        :param queue_name: The queue to bind to the exchange
        :param exchange: The source exchange to bind to
        :param routing_key: The routing key to bind on
        :param arguments: Custom key/value pair arguments for the binding

        :returns: Method frame from the Queue.Bind-ok response
        """

        return self.maintenance_channel.queue_bind(queue_name, exchange, routing_key, arguments)

    def unbind_queue(self,
                     queue_name: str,
                     exchange: str,
                     routing_key: Optional[str] = None,
                     arguments: Optional[Dict[str, Any]] = None) -> 'PikaMethod':
        """
        Unbind the queue from the specified exchange

        :param queue_name: The queue to unbind to the exchange
        :param exchange: The source exchange to unbind from
        :param routing_key: The routing key to unbind
        :param arguments: Custom key/value pair arguments for the unbinding

        :returns: Method frame from the Queue.Unbind-ok response
        """

        return self.maintenance_channel.queue_unbind(queue_name, exchange, routing_key, arguments)



import threading
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import Optional

from messageflux.base_service import BaseService, ServiceState


class _MessageFluxHTTPServer(ThreadingHTTPServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.messageflux_base_service: Optional[BaseService] = None


class WebServerLivenessAddonException(Exception):
    """
    the exception type for this addon
    """
    pass


class WebServerLivenessAddon:
    """
    an addon that adds a small webserver that listens on all routes and returns 200 if the service is alive
    or 500 otherwise.
    useful for liveness checks in deployments
    """

    def __init__(self, host: str = "0.0.0.0", port: int = 8080):
        """

        :param host: the host to listen on
        :param port: the port to listen on
        """
        self._host = host
        self._port = port
        self._service: Optional[BaseService] = None
        self._web_server: Optional[_MessageFluxHTTPServer] = None

    @property
    def service(self) -> Optional[BaseService]:
        """
        the attached service (or None if no service is attached)
        """
        return self._service

    def attach(self, service: BaseService) -> 'WebServerLivenessAddon':
        """
        attached the addon to a service, and returns the addon

        :param service: the service to attach to
        """
        if self._service is not None:
            raise WebServerLivenessAddonException('Cannot attach an already attached addon')

        if service.service_state in [ServiceState.STARTED, ServiceState.STARTING]:
            raise WebServerLivenessAddonException('Cannot monitor inactivity on an already started service')

        service.state_changed_event.subscribe(self._on_service_state_change)

        self._service = service

        return self

    def detach(self):
        """
        detaches from a service
        """
        service = self._service
        if service is None:
            return
        service.state_changed_event.unsubscribe(self._on_service_state_change)
        self._service = None
        if self._web_server is not None:
            self._web_server.messageflux_base_service = None

    def _on_service_state_change(self, service_state: ServiceState):
        if service_state == ServiceState.STARTED:
            t = threading.Thread(target=self._run_web_server,
                                 daemon=True)
            t.start()

        elif service_state == ServiceState.STOPPING:
            if self._web_server is not None:
                web_server = self._web_server
                self._web_server = None
                web_server.shutdown()

    def _run_web_server(self):
        class LivenessHandler(BaseHTTPRequestHandler):
            """
            the actual http request handler
            """

            def do_GET(self):
                assert isinstance(self.server, _MessageFluxHTTPServer)
                base_service = self.server.messageflux_base_service
                if base_service is not None and base_service.is_alive:
                    self.send_response(200)
                else:
                    self.send_response(500)
                self.end_headers()

        self._web_server = _MessageFluxHTTPServer((self._host, self._port), LivenessHandler)
        self._web_server.messageflux_base_service = self._service
        self._web_server.serve_forever()



from messageflux.iodevices.base import OutputDevice, OutputDeviceManager
from messageflux.iodevices.base.common import MessageBundle
from messageflux.iodevices.short_circuit_device_wrapper.common import ShortCircuitDeviceBase


class ShortCircuitOutputDevice(ShortCircuitDeviceBase, OutputDevice['ShortCircuitOutputDeviceManager']):
    """
    this is an output device manager that wraps input devices in short circuit input devices
    """

    def __init__(self,
                 manager: 'ShortCircuitOutputDeviceManager',
                 name: str,
                 inner_device: OutputDevice,
                 short_circuit_fail_count: int,
                 short_circuit_time: int):
        OutputDevice.__init__(self, manager, name)
        ShortCircuitDeviceBase.__init__(self, short_circuit_fail_count, short_circuit_time)
        self._inner_device = inner_device

    def _send_message(self, message_bundle: MessageBundle):
        self._validate_short_circuit()
        with self._failure_count_context():
            self._inner_device.send_message(message_bundle.message, message_bundle.device_headers)

    def close(self):
        """
        closes the inner device
        """
        self._inner_device.close()


class ShortCircuitOutputDeviceManager(OutputDeviceManager[ShortCircuitOutputDevice]):
    """
    this is an output device manager that wraps output devices in short circuit output devices
    """

    def __init__(self,
                 inner_device_manager: OutputDeviceManager,
                 short_circuit_fail_count: int,
                 short_circuit_time: int):
        self._inner_device_manager = inner_device_manager
        self._short_circuit_fail_count = short_circuit_fail_count
        self._short_circuit_fail_time = short_circuit_time

    def connect(self):
        """
        connects to the device manager
        """
        self._inner_device_manager.connect()

    def disconnect(self):
        """
        disconnects from the device manager
        """
        self._inner_device_manager.disconnect()

    def get_output_device(self, name: str) -> ShortCircuitOutputDevice:
        """
        returns a wrapped output device
        :param name: the name of the output device to get
        """
        inner_device = self._inner_device_manager.get_output_device(name)
        return ShortCircuitOutputDevice(self,
                                        name,
                                        inner_device,
                                        self._short_circuit_fail_count,
                                        self._short_circuit_fail_time)



import ssl
import threading
from abc import ABCMeta, abstractmethod
from time import perf_counter
from typing import List, Dict, Optional

from messageflux.iodevices.base import InputTransaction, ReadResult
from messageflux.iodevices.rabbitmq.rabbitmq_input_device import RabbitMQInputDevice, RabbitMQInputDeviceManager


class PoisonCounterBase(metaclass=ABCMeta):
    """
    abstract class for a persistent counter store for message ids
    """

    def start(self):
        """
        gets called before the first invocation
        """
        pass

    def stop(self):
        """
        gets called after the last invocation
        """
        pass

    @abstractmethod
    def increment_and_return_counter(self, message_id: str) -> int:
        """
        this methods increments and returns the counter for message_id.
        if there is no counter for message_id it should return '1'

        :param message_id: the message id to check the counter for
        :returns: the incremented counter for message_id
        """
        pass

    @abstractmethod
    def delete_counter(self, message_id: str):
        """
        deletes the counter for message_id

        :param message_id: the message id to delete the counter for
        """
        pass


class RabbitMQNoPoisonInputTransactionWrapper(InputTransaction):
    """
    represents a wrapper for InputTransaction for RabbitMQ with poison counter
    """

    def __init__(self,
                 inner_transaction: InputTransaction,
                 poison_counter: PoisonCounterBase,
                 message_id: str):
        """

        :param inner_transaction: the inner transaction
        :param poison_counter: the poison counter
        :param message_id: the message id in this transaction
        """
        super().__init__(inner_transaction.device)
        self._inner_transaction = inner_transaction
        self._poison_counter = poison_counter
        self._message_id = message_id

    def _commit(self):
        self._inner_transaction.commit()
        self._poison_counter.delete_counter(message_id=self._message_id)

    def _rollback(self):
        self._inner_transaction.rollback()


class RabbitMQPoisonCountingInputDevice(RabbitMQInputDevice):
    """
    represents an RabbitMQ input device, that adds handling with poison messages (for classic queues)
    """

    def __init__(self,
                 device_manager: 'RabbitMQPoisonCountingInputDeviceManager',
                 queue_name: str,
                 poison_counter: PoisonCounterBase,
                 max_poison_count=3,
                 consumer_args: Optional[Dict[str, str]] = None,
                 prefetch_count: int = 1,
                 use_consumer: bool = True):
        """
        constructs a new input RabbitMQ device

        :param device_manager: the RabbitMQ device Manager that holds this device
        :param queue_name: the name for the queue
        :param PoisonCounterBase poison_counter: the poison counter to use
        :param int max_poison_count: the number of times to requeue a message before backout
        (-1 means not to handle poison messages at all, 0 means reject all redelivered messages right away)
        :param consumer_args: the arguments to create the consumer with
        :param int prefetch_count: the number of un-acked messages that can be consumed
        :param bool use_consumer: True to use the 'consume' method, False to use 'basic_get'
        """
        super().__init__(device_manager=device_manager,
                         queue_name=queue_name,
                         consumer_args=consumer_args,
                         prefetch_count=prefetch_count,
                         use_consumer=use_consumer)
        self._max_poison_count = max_poison_count
        self._poison_counter = poison_counter

    def _get_data_from_queue(self,
                             cancellation_token: threading.Event,
                             timeout: Optional[float],
                             with_transaction: bool) -> Optional['ReadResult']:
        """
        performs a single read from queue

        :param cancellation_token: the cancellation token, to pass to transaction
        :param timeout: the timeout in seconds to block. None means no blocking
        :param with_transaction: does this read is to be done with transaction?
        :return: the stream and metadata, or None,None if no message in queue
        """
        start_time = perf_counter()

        while True:
            get_timeout: Optional[float] = None
            if timeout is not None:
                get_timeout = max(0.01, start_time + timeout - perf_counter())
            channel = self._get_channel()
            body, header_frame, method_frame = self._get_frames_from_queue(channel,
                                                                           get_timeout,
                                                                           with_transaction=with_transaction)
            if method_frame is None:  # no message in queue
                return None
            assert body is not None
            assert header_frame is not None

            redelivered = method_frame.redelivered

            if not redelivered or self._max_poison_count < 0:  # no need to protect from poison messages:
                break

            else:  # this message was redelivered. protect from poison message
                assert header_frame is not None
                message_id = f'{self.name}-{header_frame.message_id}-{header_frame.timestamp}'
                redeliver_count = self._poison_counter.increment_and_return_counter(message_id)
                if redeliver_count < self._max_poison_count:  # this message is below the max_poison_threshold
                    read_result = self._create_response_from_frames(cancellation_token=cancellation_token,
                                                                    body=body,
                                                                    header_frame=header_frame,
                                                                    method_frame=method_frame,
                                                                    channel=channel,
                                                                    with_transaction=with_transaction)
                    return ReadResult(message=read_result.message,
                                      device_headers=read_result.device_headers,
                                      transaction=RabbitMQNoPoisonInputTransactionWrapper(
                                          inner_transaction=read_result.transaction,
                                          poison_counter=self._poison_counter,
                                          message_id=message_id))

                # this point in code, means the message is a poison message.
                self._logger.error("Received message with backout count of more than {}, perhaps we are in a loop. "
                                   "handling backout. BackoutCount was {}. "
                                   "device name: {}".format(self._max_poison_count,
                                                            redeliver_count,
                                                            self._queue_name))

                self._poison_counter.delete_counter(message_id)
                if with_transaction:
                    assert method_frame is not None
                    assert method_frame.delivery_tag is not None
                    channel.basic_reject(method_frame.delivery_tag, requeue=False)

                # if the message was poison message, we want to skip it and read the next one.
                # but only if we're not exceeding timeout
                if timeout is not None:
                    if perf_counter() >= start_time + max(timeout, 1):
                        return None

                continue

        return self._create_response_from_frames(cancellation_token=cancellation_token,
                                                 body=body,
                                                 header_frame=header_frame,
                                                 method_frame=method_frame,
                                                 channel=channel,
                                                 with_transaction=with_transaction)


class RabbitMQPoisonCountingInputDeviceManager(RabbitMQInputDeviceManager):
    """
    rabbitmq input device manager, that adds handling with poison messages (for classic queues)
    """

    def __init__(self,
                 hosts: List[str],
                 user: str,
                 password: str,
                 poison_counter: PoisonCounterBase,
                 max_poison_count: int = 3,
                 port: Optional[int] = None,
                 ssl_context: Optional[ssl.SSLContext] = None,
                 virtual_host: Optional[str] = None,
                 client_args: Optional[Dict[str, str]] = None,
                 heartbeat: int = 300,
                 connection_attempts: int = 5,
                 prefetch_count: int = 1,
                 use_consumer: bool = True,
                 blocked_connection_timeout: Optional[float] = None,
                 default_direct_exchange: Optional[str] = None,
                 ):
        """
        This manager used to create RabbitMQ devices (direct queues)

        :param hosts: the list of hostnames of the manager
        :param user: the username for the rabbitMQ manager
        :param password: the password for the rabbitMQ manager
        :param PoisonCounterBase poison_counter: the poison counter to use
        :param int max_poison_count: the number of times to try to handle a message before backout
        :param port: the port to connect the hosts to
        :param ssl_context: the ssl context to use. None means don't use ssl at all
        :param virtual_host: the virtual host to connect to
        :param client_args: the arguments to create the client with
        :param int heartbeat: heartbeat interval for the connection (between 0 and 65536
        :param int connection_attempts: Maximum number of retry attempts
        (-1 means not to handle poison messages at all, 0 means reject all redelivered messages right away)
        :param int prefetch_count: the number of unacked messages that can be consumed
        :param bool use_consumer: True to use the 'consume' method, False to use 'basic_get'
        :param blocked_connection_timeout: If not None,
            the value is a non-negative timeout, in seconds, for the
            connection to remain blocked (triggered by Connection.Blocked from
            broker); if the timeout expires before connection becomes unblocked,
            the connection will be torn down, triggering the adapter-specific
            mechanism for informing client app about the closed connection:
            passing `ConnectionBlockedTimeout` exception to on_close_callback
            in asynchronous adapters or raising it in `BlockingConnection`.
        :param default_direct_exchange: optional direct exchange to bind all the queues to (None means no bind)
        """
        super().__init__(hosts=hosts,
                         user=user,
                         password=password,
                         port=port,
                         ssl_context=ssl_context,
                         virtual_host=virtual_host,
                         client_args=client_args,
                         heartbeat=heartbeat,
                         connection_attempts=connection_attempts,
                         prefetch_count=prefetch_count,
                         use_consumer=use_consumer,
                         blocked_connection_timeout=blocked_connection_timeout,
                         default_direct_exchange=default_direct_exchange)

        self._max_poison_count = max_poison_count
        self._poison_counter = poison_counter

    def connect(self):
        """
        connects to the device manager
        """
        self._poison_counter.start()
        super().connect()

    def disconnect(self):
        """
        disconnects from the device manager
        """
        self._poison_counter.stop()
        super().disconnect()

    def _device_factory(self, device_name: str) -> RabbitMQPoisonCountingInputDevice:
        return RabbitMQPoisonCountingInputDevice(device_manager=self,
                                                 queue_name=device_name,
                                                 poison_counter=self._poison_counter,
                                                 max_poison_count=self._max_poison_count,
                                                 consumer_args=self._client_args,
                                                 prefetch_count=self._prefetch_count,
                                                 use_consumer=self._use_consumer)



from abc import ABCMeta, abstractmethod
from typing import List

from messageflux.iodevices.base.common import MessageBundle
from messageflux.utils import AggregatedException


class MessageStoreException(AggregatedException):
    """
    base exception for message store exceptions
    """
    pass


class MessageStoreBase(metaclass=ABCMeta):
    """
    base class for a Message Store
    """

    @property
    @abstractmethod
    def magic(self) -> bytes:
        """
        return a magic prefix that is unique and constant for this message store
        """
        pass

    def connect(self):
        """
        connects to Message Store

        It is expected, that the implementation stays connected to the message store since 'connect' is called, and
        up until 'close' is called.

        if the for some reason the connection terminates, the implementation must try to re-connect silently upon
        operation
        """
        pass

    def disconnect(self):
        """
        closes the connection to Message Store
        """
        pass

    def __enter__(self) -> 'MessageStoreBase':
        """
        enters the context for this Message Store

        :return: self
        """
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        exits the context for this Message Store (calls close)

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        """
        self.disconnect()

    @abstractmethod
    def read_message(self, key: str) -> MessageBundle:
        """
        reads a message according to the key given

        :param str key: the key to the message
        :return: a Message from the store
        """
        pass

    @abstractmethod
    def put_message(self, device_name: str, message_bundle: MessageBundle) -> str:
        """
        puts a message in the message store

        :param device_name: the name of the device putting the item in the store
        :param message_bundle: the Message bundle to write to the store
        :return: the key to the message in the message store
        """
        pass

    @abstractmethod
    def delete_message(self, key: str):
        """
        deletes a message from the message store

        :param str key: the key to the message
        """
        pass

    def delete_messages(self, keys: List[str]):
        """
        deletes multiple messages from the message store

        :param list[str] keys: the list of keys to the messages
        """
        failures = []

        for key in keys:
            try:
                self.delete_message(key)
            except Exception as ex:
                failures.append(ex)

        if failures:
            raise MessageStoreException(
                "Error deleting {} out of {} messages from store".format(len(failures), len(keys)),
                inner_exceptions=failures)



import os
import random
import socket
import threading
import traceback
from abc import ABCMeta, abstractmethod
from datetime import datetime
from logging.handlers import BaseRotatingHandler
from os import scandir
from shutil import move
from typing import Optional, Iterator

import itertools
from time import sleep, time

from messageflux.utils import get_random_id


class BulkRotatingHandlerBase(BaseRotatingHandler, metaclass=ABCMeta):
    """
    Handler for logging to a destination, while using a rotating "live" file, which rotates "into" its destination
    when the current file reaches a certain size or time.
    """
    _MIN_SLEEP_TIME = 5

    def __init__(self,
                 live_log_path: str,
                 bkp_log_path: Optional[str] = None,
                 max_records: int = 1000,
                 max_time: int = 60,
                 live_log_prefix: str = ''):
        """
        Open a log file and use it as the stream for logging.
        when max_records are written, or max_time has passed, a rollover occurs
        rollover copies the live log from live_log_path to the "destination"

        :param live_log_path: the path to write the live log to
        :param bkp_log_path: the path to write to rotated log to, if writing to rotated_log_path fails
        :param max_records: the maximum number of records to write before rotation
        :param max_time: the maximum time (in seconds) to wait before rotation
        :param live_log_prefix: the prefix for live log file
        """
        self._live_log_path = os.path.abspath(live_log_path)
        filename = os.path.join(self._live_log_path,
                                f'{live_log_prefix}{socket.gethostname()}-{str(os.getpid())}-{id(self)}.log')
        os.makedirs(self._live_log_path, exist_ok=True)

        BaseRotatingHandler.__init__(self, filename, 'a', None, True)

        if bkp_log_path is None:
            self._bkp_log_path = self._live_log_path
        else:
            self._bkp_log_path = os.path.abspath(bkp_log_path)
            os.makedirs(self._bkp_log_path, exist_ok=True)

        self._max_records = max_records
        self._max_time = max_time
        self._record_count = 0
        self._run = True
        self._next_rotate = time() + self._max_time
        self._log_thread: Optional[threading.Thread] = None
        self.doRollover()
        self._bkp_copy_thread = threading.Thread(target=self._do_bkp_copy_thread, daemon=True)
        self._bkp_copy_thread.start()

    @abstractmethod
    def _move_log_to_destination(self, src_file):
        """
        this moves the live log from a file, to its destination
        """
        raise NotImplementedError()

    @staticmethod
    def _safe_move(src, dst):
        """
        moves a file from src to dst as atomically possible

        :param src: src filepath
        :param dst: dsf filepath
        """
        try:
            os.chmod(src, 0o777)
            os.rename(src, dst)  # Try to rename - most atomic possible
        except Exception:
            basedir = os.path.dirname(dst)
            tmpdir = os.path.join(basedir, '_TMP_')
            os.makedirs(tmpdir, exist_ok=True)  # create a tmp directory under the destination directory
            tmpfile = os.path.join(tmpdir, "{}.tmp".format(get_random_id()))
            move(src, tmpfile)  # move the src to tmp dir in dest directory
            move(tmpfile, dst)  # move the tmp file to dest file (now they're in the same filesystem: should be atomic)

    @staticmethod
    def _get_unique_log_filename() -> str:
        """
        generates a random filename for log

        :return: a random filename for log
        """
        return "{host}_{time}_{rnd}_{pid}.log".format(
            host=socket.gethostname(),
            time=datetime.now().strftime('%y_%m_%d_%H_%M_%S_%f'),
            rnd=random.randint(0, 100000),
            pid=os.getpid()
        )

    @staticmethod
    def _sleep_until(timestamp):
        now = time()
        delay = timestamp - now
        if delay > 0:
            sleep(delay)

    def _do_logger_rotate_thread(self):
        while self._run:
            try:
                if self.shouldRollover(None):
                    # flush all bulks
                    self.doRollover()
            except Exception:
                print('Error in Logger thread. error:\r\n' + traceback.format_exc())

            self._sleep_until(self._next_rotate)

    def _do_bkp_copy_thread(self):
        while self._run:
            try:
                now = time()
                entries: Iterator[os.DirEntry] = scandir(self._bkp_log_path)
                if self._live_log_path != self._bkp_log_path:
                    entries = itertools.chain(entries, scandir(self._live_log_path))

                for direntry in entries:
                    if not direntry.is_file():
                        continue
                    src = direntry.path
                    if src == self.baseFilename:
                        # don't move the file we use to log right now
                        continue
                    file_age = now - direntry.stat().st_mtime
                    if file_age < self._max_time * 2:
                        # don't move files that are too new (might be a file of another pod/process)
                        continue
                    self._move_log_to_destination(src)
            except Exception:
                print('Error in bkp copy thread. error:\r\n' + traceback.format_exc())
            # sleep for at least a few seconds, so we won't spam
            sleep(max(self._max_time,
                      self._MIN_SLEEP_TIME))  # could be anything - but max_is a good estimate of how long is allowed

    def _ensure_logger_rotate_thread(self):
        """
        ensures that the logger thread starts once if needed
        :return:
        """
        if self._log_thread is None and self._run:
            self._log_thread = threading.Thread(target=self._do_logger_rotate_thread, daemon=True)
            self._log_thread.start()

    def emit(self, record):
        self._ensure_logger_rotate_thread()
        super(BulkRotatingHandlerBase, self).emit(record)
        self._record_count += 1

    def doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        self.acquire()
        try:
            if self.stream:
                self.stream.close()
                self.stream = None  # type: ignore

            if os.path.exists(self.baseFilename):
                try:
                    self._move_log_to_destination(self.baseFilename)
                except Exception:
                    dst_filename = os.path.join(self._bkp_log_path, self._get_unique_log_filename())
                    self._safe_move(self.baseFilename, dst_filename)
        finally:
            self._record_count = 0
            self._next_rotate = time() + self._max_time
            self.release()

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.

        Basically, see if the supplied record would cause the file to exceed
        the size limit we have.
        """

        if self._record_count >= self._max_records:
            return True

        if time() >= self._next_rotate:
            return True

        return False

    def close(self):
        """
        stops the logger thread and flushes
        """
        super(BulkRotatingHandlerBase, self).close()
        self._run = False
        self.doRollover()



import threading
from typing import Optional

from messageflux.iodevices.base import InputDevice, ReadResult, InputDeviceManager
from messageflux.iodevices.short_circuit_device_wrapper.common import ShortCircuitDeviceBase


class ShortCircuitInputDevice(ShortCircuitDeviceBase, InputDevice['ShortCircuitInputDeviceManager']):
    """
    this is a wrapper input device for adding the short circuit logic to input devices
    """

    def __init__(self,
                 manager: 'ShortCircuitInputDeviceManager',
                 name: str,
                 inner_device: InputDevice,
                 short_circuit_fail_count: int,
                 short_circuit_time: int):
        InputDevice.__init__(self, manager, name)
        ShortCircuitDeviceBase.__init__(self, short_circuit_fail_count, short_circuit_time)
        self._inner_device = inner_device

    def _read_message(self,
                      cancellation_token: threading.Event,
                      timeout: Optional[float] = None,
                      with_transaction: bool = True) -> Optional[ReadResult]:
        self._validate_short_circuit()
        with self._failure_count_context():
            return self._inner_device.read_message(cancellation_token=cancellation_token,
                                                   timeout=timeout,
                                                   with_transaction=with_transaction)

    def close(self):
        """
        closes the inner device
        """
        self._inner_device.close()


class ShortCircuitInputDeviceManager(InputDeviceManager[ShortCircuitInputDevice]):
    """
    this is an input device manager that wraps input devices in short circuit input devices
    """

    def __init__(self,
                 inner_device_manager: InputDeviceManager,
                 short_circuit_fail_count: int,
                 short_circuit_time: int):
        """

        :param inner_device_manager: the inner device manager
        :param short_circuit_fail_count: the number of consecutive failures after which,
        the device will be short circuited
        :param short_circuit_time: the time that the device will remain short circuited
        """
        self._inner_device_manager = inner_device_manager
        self._short_circuit_fail_count = short_circuit_fail_count
        self._short_circuit_fail_time = short_circuit_time

    def connect(self):
        """
        connects to the device manager
        """
        self._inner_device_manager.connect()

    def disconnect(self):
        """
        disconnects from the device manager
        """
        self._inner_device_manager.disconnect()

    def get_input_device(self, name: str) -> ShortCircuitInputDevice:
        """
        returns a wrapped input device

        :param name: the name of the input device to get
        """
        inner_device = self._inner_device_manager.get_input_device(name)
        return ShortCircuitInputDevice(self,
                                       name,
                                       inner_device,
                                       self._short_circuit_fail_count,
                                       self._short_circuit_fail_time)



import logging
from typing import Collection, List, Callable

from messageflux.iodevices.base import OutputDeviceManager, OutputDevice, OutputDeviceException
from messageflux.iodevices.base.common import MessageBundle


class CollectionOutputDevice(OutputDevice['CollectionOutputDeviceManager']):
    """
    this class writes to a collection of output devices.

    it iterates over the collection, and sends to the first successful device
    """

    def __init__(self,
                 device_manager: 'CollectionOutputDeviceManager',
                 device_name: str,
                 output_devices: Collection[OutputDevice]):
        """
        :param device_name: the name of this device
        :param output_devices: the outgoing devices to send to
        """
        super(CollectionOutputDevice, self).__init__(device_manager, device_name)
        self._output_devices = output_devices

        self._logger = logging.getLogger(__name__)

    def _send_message(self, message_bundle: MessageBundle):
        """
        sends a message to the device. this should be implemented by child classes

        :param message_bundle: the message bundle to send
        """
        failures = []
        for curr_device in self._output_devices:
            try:
                curr_device.send_message(message_bundle.message, message_bundle.device_headers)
                return
            except Exception as e:
                failures.append(e)
                self._logger.warning(f'underlying device {curr_device.name}', exc_info=True)

        self._logger.error("Couldn't send to any underlying device")
        raise OutputDeviceException("Couldn't send to any underlying device", inner_exceptions=failures)

    def close(self):
        """
        closes the connection to device
        """
        for device in self._output_devices:
            try:
                device.close()
            except Exception:
                self._logger.warning(f'Error closing underlying device {device.name}', exc_info=True)


class CollectionOutputDeviceManager(OutputDeviceManager[CollectionOutputDevice]):
    """
    This class is used to create Collection OutputDevices
    """

    def __init__(self,
                 inner_managers: List[OutputDeviceManager],
                 collection_maker: Callable[[List[OutputDevice]], Collection[OutputDevice]]):
        """
        This class is used to create Collection OutputDevices

        :param inner_managers: the actual OutputDeviceManager instances to generate devices from
        :param collection_maker: the callable to make the iterable collection from list of devices
        """
        self._inner_managers = inner_managers
        self._logger = logging.getLogger(__name__)
        self._collection_maker = collection_maker

    def connect(self):
        """
        connects to device manager
        """
        failures = []
        for manager in self._inner_managers:
            try:
                manager.connect()
            except Exception as e:
                failures.append(e)
                self._logger.warning(
                    f'Error connecting to underlying manager: {type(manager).__name__}', exc_info=True)

        if len(failures) >= len(self._inner_managers):
            self._logger.error("Couldn't connect to any underlying manager")
            raise OutputDeviceException('Error connecting to underlying device managers', inner_exceptions=failures)

    def disconnect(self):
        """
        closes the connection to IODeviceManager
        """
        for manager in self._inner_managers:
            try:
                manager.disconnect()
            except Exception:
                self._logger.warning(
                    f'Error closing underlying manager {type(manager).__name__}', exc_info=True)

    def get_output_device(self, name: str) -> CollectionOutputDevice:
        """
        Returns an output device by name

        :param name: the name of the device to write to
        :return: an output device for 'device_name'
        """
        devices: List[OutputDevice] = []
        failures = []
        for manager in self._inner_managers:
            try:
                device = manager.get_output_device(name)
                devices.append(device)
            except Exception as ex:
                failures.append(ex)
                self._logger.warning(
                    f"Error creating output device {name} from {type(manager).__name__} manager.", exc_info=True)
        if not devices:
            self._logger.error(f"Couldn't create output device '{name}'")
            raise OutputDeviceException(f"Couldn't create output device '{name}'", inner_exceptions=failures)

        return CollectionOutputDevice(self, name, self._collection_maker(devices))



import json
import os
import re
from abc import abstractmethod, ABCMeta
from hashlib import md5
from typing import Optional, BinaryIO, Dict, Any, Tuple, TYPE_CHECKING

from messageflux.iodevices.base.common import MessageBundle, Message
from messageflux.iodevices.message_store_device_wrapper.message_store_base import MessageStoreException, \
    MessageStoreBase
from messageflux.iodevices.objectstorage.s3api.s3bucket import BUCKET_NAME_VALIDATOR, S3Bucket
from messageflux.metadata_headers import MetadataHeaders
from messageflux.utils import get_random_id, json_safe_encoder

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3ServiceResource

KEY_HEADER_CONST = "__KEY__"

_S3_TIMEOUT = int(os.environ.get("S3_TIMEOUT", 1))
_S3_RETRIES = int(os.environ.get("S3_RETRIES", 2))


class BucketNameFormatterBase:
    """
    a base class for formatter to manipulate the bucket name in case it needs to be different then the device name
    """
    BUCKET_SANITATION_RE = re.compile(r'[^a-z0-9.\-]')

    def _sanitize_bucket_name(self, bucket_name: str) -> str:
        sanitized_bucket_name = self.BUCKET_SANITATION_RE.sub('-', bucket_name.lower())
        # Some devices have an underscore ("_") in their name which is not allowed for S3Bucket. So replace it with "-"
        if not BUCKET_NAME_VALIDATOR.match(sanitized_bucket_name):
            raise MessageStoreException(
                'Invalid bucket name. Bucket name must be between 3 and 63 characters long, and '
                'contain only lowercase letters, digits, dots (.) and hyphens(-). '
                f'Bucket name was: {bucket_name}')
        return sanitized_bucket_name

    def _inner_format_name(self, device_name: str, message_bundle: MessageBundle) -> str:
        """
        uses the given parameters to determine what the bucket name should be.
        this should be inherited and implemented in children

        :param device_name: the name of the device
        :param message_bundle: the message that was sent to the device
        :return: the bucket name
        """
        return device_name

    def format_name(self, device_name: str, message_bundle: MessageBundle) -> str:
        """
        uses the given parameters to determine what the bucket name should be

        :param device_name: the name of the device
        :param message_bundle: the message that was sent to the device
        :return: the bucket name
        """
        formatted_name = self._inner_format_name(device_name, message_bundle)
        return self._sanitize_bucket_name(formatted_name)


class _S3MessageStoreBase(MessageStoreBase, metaclass=ABCMeta):
    """
    a message store that uses S3 as it's base
    """
    _ORIGINAL_HEADERS_KEY = "originalheaders"

    def __init__(self,
                 s3_resource: 'S3ServiceResource',
                 magic: bytes,
                 auto_create_bucket: bool = False,
                 bucket_name_formatter: Optional[BucketNameFormatterBase] = None):
        """
        An S3 based message store

        :param s3_resource: the s3 resource from boto
        :param auto_create_bucket: Whether or not a bucket will be created
                                   when a message is being put in a nonexistent one.
        :param bucket_name_formatter: a formatter to use to manipulate the bucket name.
                                      if none is given the device name will be used
        """
        self.bucket_name_formatter = bucket_name_formatter or BucketNameFormatterBase()
        self._magic = magic

        self._s3_resource = s3_resource
        self._auto_create_bucket = auto_create_bucket
        self._bucket_cache: Dict[str, S3Bucket] = {}

    def _get_bucket(self, bucket_name: str, auto_create=False) -> S3Bucket:
        """
        gets the bucket from cache
        """
        bucket = self._bucket_cache.get(bucket_name, None)
        if bucket is None:
            bucket = S3Bucket(bucket_name, self._s3_resource, auto_create=auto_create)
            self._bucket_cache[bucket_name] = bucket

        return bucket

    @property
    def magic(self) -> bytes:
        """
        return a magic string that is unique and constant for this message store
        """
        return self._magic

    def _serialize_key(self, bucket: S3Bucket, key: str) -> str:
        """
        serializes the key for sending on the wire

        :param bucket: the bucket
        :param key: the key in the bucket
        :return: serialized key to send
        """
        data_dict = {'bucket_name': bucket.name, 'key': key}
        return json.dumps(data_dict)

    def deserialize_key(self, data: str) -> Tuple[str, str]:
        """
        deserializes the key and bucket received from  the wire

        :param data: the data from the wire
        :return: deserialized (bucket name, key)
        """
        data_dict: Dict[str, Any] = json.loads(data)
        bucket_name = str(data_dict['bucket_name'])
        key = str(data_dict['key'])
        return bucket_name, key

    def _serialize_headers(self, headers: Dict[str, Any]) -> Dict[str, str]:
        return {self._ORIGINAL_HEADERS_KEY: json.dumps(headers, default=json_safe_encoder)}

    def _deserialize_headers(self, headers: Dict[str, str]) -> Dict[str, Any]:
        return json.loads(headers.get(self._ORIGINAL_HEADERS_KEY, '{}'))

    @abstractmethod
    def _read_message_from_bucket(self, bucket: S3Bucket, key: str) -> Tuple[BinaryIO, Dict[str, Any]]:
        """
        reads the message from the current s3 api
        """
        pass

    @abstractmethod
    def _put_message_to_bucket(self, bucket: S3Bucket, key: str, message: BinaryIO, metadata: Dict[str, str]):
        pass

    def read_message(self, key: str) -> MessageBundle:
        """
        reads a message according to the key given
        :return: a tuple of the bytes of the message to read, and its metadata
        """
        bucket_name, s3_key = self.deserialize_key(key)
        bucket = self._get_bucket(bucket_name)
        body, headers = self._read_message_from_bucket(bucket, s3_key)

        message_headers = self._deserialize_headers(headers)
        device_headers = {KEY_HEADER_CONST: s3_key}

        return MessageBundle(Message(body, message_headers), device_headers)

    def put_message(self, device_name: str, message_bundle: MessageBundle) -> str:
        """
        puts a message in the message store

        :param device_name: the name of the device putting the item in the store
        :param message_bundle: the Message bundle to write to the store
        :return: the key to the message in the message store
        """
        bucket_name = self.bucket_name_formatter.format_name(device_name, message_bundle)
        bucket = self._get_bucket(bucket_name=bucket_name, auto_create=self._auto_create_bucket)
        data_hash = md5(message_bundle.message.bytes).hexdigest()
        key = \
            message_bundle.message.headers.get(KEY_HEADER_CONST) or \
            message_bundle.device_headers.get(MetadataHeaders.ITEM_ID) or \
            get_random_id()

        key = key + '.' + data_hash
        serialized_headers = self._serialize_headers(message_bundle.message.headers)
        self._put_message_to_bucket(bucket=bucket,
                                    key=key,
                                    message=message_bundle.message.stream,
                                    metadata=serialized_headers)
        return self._serialize_key(bucket, key)

    def delete_message(self, key: str):
        """
        deletes a message from the message store
        :param str key: the key to the message
        """
        bucket_name, s3_key = self.deserialize_key(key)
        bucket = self._get_bucket(bucket_name=bucket_name)
        bucket.delete_object(s3_key)


class S3MessageStore(_S3MessageStoreBase):
    """
    a message store that uses S3 as it's base
    """

    def __init__(self,
                 s3_resource: 'S3ServiceResource',
                 magic: bytes = b"__S3_MSGSTORE__",
                 auto_create_bucket: bool = False,
                 bucket_name_formatter: Optional[BucketNameFormatterBase] = None,
                 put_object_extra_args: Optional[Dict[str, Any]] = None):
        """
        An S3 based message store

        :param s3_resource: the s3 resource from boto
        :param auto_create_bucket: Whether or not a bucket will be created
                                   when a message is being put in a nonexistent one.
        :param bucket_name_formatter: a formatter to use to manipulate the bucket name.
                                      if none is given the device name will be used
        :param put_object_extra_args: extra args to give to bucket.put_object(). i.e 'StorageClass'
        """
        super().__init__(s3_resource=s3_resource,
                         magic=magic,
                         auto_create_bucket=auto_create_bucket,
                         bucket_name_formatter=bucket_name_formatter)
        self._put_object_extra_args = put_object_extra_args or {}

    def _read_message_from_bucket(self, bucket: S3Bucket, key: str) -> Tuple[BinaryIO, Dict[str, Any]]:
        s3obj = bucket.get_object(key=key)
        headers = s3obj.metadata
        return s3obj.body, headers

    def _put_message_to_bucket(self, bucket: S3Bucket, key: str, message: BinaryIO, metadata: Dict[str, str]):
        bucket.put_object(key=key,
                          buf=message,
                          metadata=metadata,
                          **self._put_object_extra_args)


class S3UploadMessageStore(_S3MessageStoreBase):
    """
    a message store that uses S3 as it's base (using upload_file method)
    """

    def __init__(self,
                 s3_resource: 'S3ServiceResource',
                 magic: bytes = b"__S3_UPLOAD_MSGSTORE__",
                 auto_create_bucket: bool = False,
                 bucket_name_formatter: Optional[BucketNameFormatterBase] = None,
                 upload_extra_args: Optional[Dict[str, Any]] = None):
        """
        An S3 based message store

        :param s3_resource: the s3 resource from boto
        :param auto_create_bucket: Whether or not a bucket will be created
                                   when a message is being put in a nonexistent one.
        :param bucket_name_formatter: a formatter to use to manipulate the bucket name.
                                      if none is given the device name will be used
        :param upload_extra_args: extra args to give to client.upload_fileobj(). i.e 'StorageClass'
        """
        super().__init__(s3_resource=s3_resource,
                         magic=magic,
                         auto_create_bucket=auto_create_bucket,
                         bucket_name_formatter=bucket_name_formatter)
        self._upload_extra_args = upload_extra_args or {}

    def _read_message_from_bucket(self, bucket: S3Bucket, key: str) -> Tuple[BinaryIO, Dict[str, Any]]:
        s3obj = bucket.get_object(key=key)
        headers = s3obj.metadata
        return s3obj.body, headers

    def _put_message_to_bucket(self, bucket: S3Bucket, key: str, message: BinaryIO, metadata: Dict[str, str]):
        bucket.upload_object(key=key,
                             stream=message,
                             metadata=metadata,
                             **self._upload_extra_args)



import os
import socket
from os import scandir
from typing import List, Optional

from messageflux.iodevices.file_system.file_system_serializer import FileSystemSerializerBase, \
    DefaultFileSystemSerializer
from messageflux.utils import get_random_id, KwargsException


class FileSystemDeviceManagerBase:
    """
    this manager is used to create IO Devices over a shared file system
    """
    DEFAULT_QUEUES_SUB_DIR = 'QUEUES'
    DEFAULT_BOOKKEEPING_SUB_DIR = 'BOOKKEEPING'
    DEFAULT_TMPDIR_SUB_DIR = 'TMP'

    def __init__(self,
                 root_folder: str,
                 queue_dir_name: str = DEFAULT_QUEUES_SUB_DIR,
                 tmp_dir_name: str = DEFAULT_TMPDIR_SUB_DIR,
                 bookkeeping_dir_name: str = DEFAULT_BOOKKEEPING_SUB_DIR,
                 serializer: Optional[FileSystemSerializerBase] = None):
        """
        :param root_folder: the root folder to use for the manager
        :param queue_dir_name: the name of the subdirectory under root_folder that holds the queues
        :param tmp_dir_name: the name of the subdirectory under root_folder to use for temp files
        :param bookkeeping_dir_name: the name of the subdirectory under root_folder that holds the book-keeping data
        :param serializer: the serializer to use to write messages to files. None will use the default serializer
        """
        self._root_folder = root_folder
        self._queues_folder = os.path.join(root_folder, queue_dir_name)
        self._tmp_folder = os.path.join(root_folder, tmp_dir_name)
        self._bookkeeping_folder = os.path.join(root_folder, bookkeeping_dir_name)
        self._unique_manager_id = f'{socket.gethostname()}-{get_random_id()}'
        self._serializer = serializer or DefaultFileSystemSerializer()

    @property
    def unique_manager_id(self) -> str:
        """
        this is a unique id for this manager instance
        """
        return self._unique_manager_id

    @property
    def bookkeeping_folder(self) -> str:
        """
        the full path for the book keeping folder
        """
        return self._bookkeeping_folder

    @property
    def queues_folder(self) -> str:
        """
        the full path for the queues folder
        """
        return self._queues_folder

    @property
    def tmp_folder(self) -> str:
        """
        the full path for the temp folder
        """
        return self._tmp_folder

    def get_available_device_names(self) -> List[str]:
        """
        returns a list of available 'queues' in the queue folder
        """
        available_devices = []
        for dir_entry in scandir(self._queues_folder):
            if dir_entry.is_dir():
                available_devices.append(dir_entry.name)
        return available_devices

    def _create_all_directories(self):
        """
        connects to device manager
        """
        try:
            os.makedirs(self._tmp_folder, exist_ok=True)
            os.makedirs(self._queues_folder, exist_ok=True)
            os.makedirs(self._bookkeeping_folder, exist_ok=True)
        except Exception as e:
            raise KwargsException('Error creating directories') from e



from .common import (Message,
                     MessageHeaders,
                     MessageBundle,
                     DeviceHeaders)

from .input_devices import (InputDevice,
                            InputDeviceManager,
                            InputDeviceException,
                            ReadResult,
                            AggregatedInputDevice,
                            NULL_TRANSACTION)

from .input_transaction import (InputTransaction,
                                InputTransactionScope)

from .output_devices import (OutputDevice,
                             OutputDeviceManager,
                             OutputDeviceException)



import logging

from messageflux.iodevices.base import InputTransaction, InputDevice, ReadResult, InputDeviceManager
from messageflux.iodevices.message_store_device_wrapper.message_store_base import MessageStoreBase
from messageflux.iodevices.message_store_device_wrapper.message_store_transformer_base import \
    _MessageStoreTransformerBase
from messageflux.iodevices.transformer_device_wrapper import InputTransformerBase, TransformerInputDeviceManager
from messageflux.iodevices.transformer_device_wrapper.transformer_input_device import TransformerInputDevice


class MessageStoreInputTransactionWrapper(InputTransaction):
    """
    represents an Message Store wrapper transaction
    """

    _logger = logging.getLogger(__name__)

    def __init__(self,
                 device: InputDevice,
                 inner_transaction: InputTransaction,
                 message_store: MessageStoreBase,
                 key: str,
                 delete_on_commit: bool = True):
        """
        represents an message store wrapper transaction

        :param device: the input device that returned that transaction
        :param inner_transaction: the actual transaction
        :param MessageStoreBase message_store: the message store that stores this item
        :param str key: the key to this message
        :param delete_on_commit: Whether or not we should delete the stored buffer after committing a message from the
                                 relevant input device
        """
        super(MessageStoreInputTransactionWrapper, self).__init__(device)
        self._inner_transaction = inner_transaction
        self._delete_on_commit = delete_on_commit
        self._message_store = message_store
        self._key = key

    def _commit(self):
        """
        commits the transaction
        """
        self._inner_transaction.commit()
        if self._delete_on_commit:
            try:
                self._message_store.delete_message(key=self._key)
            except Exception:
                self._logger.warning(f'Error deleting key {self._key}', exc_info=True)

    def _rollback(self):
        """
        rolls back the transaction
        """
        self._inner_transaction.rollback()


class MessageStoreInputTransformer(_MessageStoreTransformerBase, InputTransformerBase):
    """
    a transformer class that converts key to message from the message store
    """

    def __init__(self,
                 message_store: MessageStoreBase,
                 delete_on_commit: bool = True):
        """
        :param message_store: the message store to use
        :param delete_on_commit: should we the delete the message from the message store on transaction commit?
        """
        super(MessageStoreInputTransformer, self).__init__(message_store=message_store)
        self._delete_on_commit = delete_on_commit

    def transform_incoming_message(self, input_device: TransformerInputDevice, read_result: ReadResult) -> ReadResult:
        """
        Transform the message that was received from the underlying device.
        in our case, converts the key received from the device, to the actual message from the message store
        in case the message received is not a key, doesn't do anything

        :param input_device: the input device that the transformer runs on
        :param read_result: the original ReadResult received from the underlying device
        :return: the transformed ReadResult
        """
        try:
            if not self.is_key(read_result.message.stream):
                return read_result

            key = self.deserialize_key(read_result.message.bytes)
            store_message_bundle = self._message_store.read_message(key)

            wrapped_transaction = MessageStoreInputTransactionWrapper(input_device,
                                                                      inner_transaction=read_result.transaction,
                                                                      message_store=self._message_store,
                                                                      key=key,
                                                                      delete_on_commit=self._delete_on_commit)

            store_message_bundle.message.headers.update(read_result.message.headers)
            store_message_bundle.device_headers.update(read_result.device_headers)

        except Exception:
            read_result.rollback()
            raise

        return ReadResult(message=store_message_bundle.message,
                          device_headers=store_message_bundle.device_headers,
                          transaction=wrapped_transaction)


class MessageStoreInputDeviceManagerWrapper(TransformerInputDeviceManager):
    """
    This class is used to wrap InputDeviceManager with message store functionality
    """

    def __init__(self,
                 inner_manager: InputDeviceManager,
                 message_store: MessageStoreBase,
                 delete_on_commit: bool = True):
        """
        This class is used to wrap IODeviceManager with message store functionality

        :param inner_manager: the actual manager to wrap
        :param message_store: the message store to use to store files
        :param delete_on_commit: When reading a message that was stored in the relevant message store, should we delete
                                 the stored message when committing the original input message? Sometimes we might want
                                 to leave the stored message in order to avoid conflicts between systems that read
                                 duplications of the same message.
        """
        transformer = MessageStoreInputTransformer(message_store=message_store,
                                                   delete_on_commit=delete_on_commit)
        super(MessageStoreInputDeviceManagerWrapper, self).__init__(inner_manager, transformer)



from messageflux.iodevices.base import InputDevice, ReadResult

from .message_handling_service import (BatchMessageHandlingService,
                                       MessageHandlerBase,
                                       MessageHandlingService,
                                       BatchMessageHandlerBase)

from .pipeline_service import (PipelineService,
                               PipelineResult,
                               PipelineHandlerBase)



from random import shuffle
from typing import List

from messageflux.iodevices.base import InputDeviceManager, OutputDeviceManager
from messageflux.iodevices.collection_device_wrapper import CollectionInputDeviceManager, CollectionOutputDeviceManager
from messageflux.utils import StatefulListIterator


def _create_round_robin_collection(devices: List):
    shuffle(devices)
    return StatefulListIterator(devices)


class RoundRobinInputDeviceManager(CollectionInputDeviceManager):
    """
    This class is used to create RoundRobin InputDevices
    """

    def __init__(self, inner_managers: List[InputDeviceManager]):
        """
        This class is used to create RoundRobin IODevices

        :param inner_managers: the actual InputDeviceManager instances to generate devices from
        """

        super(RoundRobinInputDeviceManager, self).__init__(inner_managers=inner_managers,
                                                           collection_maker=_create_round_robin_collection)


class RoundRobinOutputDeviceManager(CollectionOutputDeviceManager):
    """
    This class is used to create RoundRobin IODevices
    """

    def __init__(self, inner_managers: List[OutputDeviceManager]):
        """
        This class is used to create RoundRobin IODevices

        :param inner_managers: the actual OutputDeviceManager instances to generate devices from
        """
        super(RoundRobinOutputDeviceManager, self).__init__(inner_managers=inner_managers,
                                                            collection_maker=_create_round_robin_collection)



from messageflux.iodevices.base import OutputDeviceException, OutputDeviceManager
from messageflux.iodevices.base.common import MessageBundle, Message
from messageflux.iodevices.message_store_device_wrapper.message_store_base import MessageStoreBase
from messageflux.iodevices.message_store_device_wrapper.message_store_transformer_base import \
    _MessageStoreTransformerBase
from messageflux.iodevices.transformer_device_wrapper import OutputTransformerBase
from messageflux.iodevices.transformer_device_wrapper.transformer_output_device import TransformerOutputDevice, \
    TransformerOutputDeviceManager


class MessageStoreOutputTransformer(_MessageStoreTransformerBase, OutputTransformerBase):
    """
    a transformer class that converts key to message from the message store
    """

    ORIGINAL_MESSAGE_SIZE_HEADER = '__ORIGINAL_MESSAGE_SIZE_HEADER__'

    def __init__(self,
                 message_store: MessageStoreBase,
                 size_threshold: int = -1):
        """
        :param message_store: the message store to use
        :param size_threshold: the minimum message size, the only above we store the message in message store.
        """
        super().__init__(message_store=message_store)
        self._size_threshold = size_threshold

    def transform_outgoing_message(self,
                                   output_device: TransformerOutputDevice,
                                   message_bundle: MessageBundle) -> MessageBundle:
        """
        Transform the message before it is sent to the underlying device.
        in our case, stores the message in message store, and returns the key as the output message

        if the message size is under size_threshold, nothing is done

        :param output_device: the output device that the transformer runs on
        :param message_bundle: the message bundle to transform
        :return: the transformed message bundle to send through the underlying device
        """

        stream_size = message_bundle.message.stream.seek(0, 2)
        message_bundle.message.stream.seek(0)

        if stream_size > self._size_threshold:
            # send using massage store
            result_message = self._store_in_message_store(output_device, message_bundle)
            result_message.headers[self.ORIGINAL_MESSAGE_SIZE_HEADER] = stream_size
            return MessageBundle(result_message, message_bundle.device_headers)
        else:
            # send without message store
            return message_bundle

    def _store_in_message_store(self,
                                output_device: TransformerOutputDevice,
                                message_bundle: MessageBundle) -> Message:
        """
        stores the message in the message store

        :param output_device: the output device that the transformer runs on
        :param message_bundle: the message bundle to store

        :return: the message containing the key to the message store
        """
        try:
            # save to object storage and get the key
            key = self._message_store.put_message(output_device.name, message_bundle)
        except Exception as ex:
            raise OutputDeviceException('Error putting item into message store') from ex
        try:
            # send the key instead of the buffer over the inner device
            data = self.serialize_key(key)
            return Message(data, message_bundle.message.headers)
        except Exception:
            try:
                # delete from the object storage
                self._message_store.delete_message(key)
            except Exception as del_ex:
                raise OutputDeviceException(
                    'Error deleting item from message store') from del_ex  # ex will be in __context__
            raise


class MessageStoreOutputDeviceManagerWrapper(TransformerOutputDeviceManager):
    """
    This class is used to wrap IODeviceManager with message store functionality
    """

    def __init__(self,
                 inner_manager: OutputDeviceManager,
                 message_store: MessageStoreBase,
                 size_threshold: int = -1):
        """
        This class is used to wrap IODeviceManager with message store functionality

        :param inner_manager: the actual manager to wrap
        :param MessageStoreBase message_store: the message store to use to store files
        :param size_threshold: only use message store if the size of the product (in bytes) is larger than this.
                               set to -1 to always use message store (default: -1).

        """
        transformer = MessageStoreOutputTransformer(message_store=message_store,
                                                    size_threshold=size_threshold)

        super(MessageStoreOutputDeviceManagerWrapper, self).__init__(inner_manager, transformer)
        self.size_threshold = size_threshold



import json
import zipfile
from abc import ABCMeta, abstractmethod
from io import BytesIO
from typing import BinaryIO

from messageflux.iodevices.base import Message
from messageflux.utils import json_safe_encoder


class FileSystemSerializerBase(metaclass=ABCMeta):
    """
    this is a base class for FileSystemSerializer
    """

    @abstractmethod
    def serialize(self, message: Message) -> BinaryIO:
        """
        serializes the message into a stream to write to file

        :param message: the message to serialize
        :return: a stream containing the serialized message
        """
        raise NotImplementedError()

    @abstractmethod
    def deserialize(self, stream: BinaryIO) -> Message:
        """
        deserializes the message from the stream

        :param stream: the stream to deserialize from
        :return: the deserialized message
        """
        raise NotImplementedError()


class ConcatFileSystemSerializer(FileSystemSerializerBase):
    """
    this serializer concats the headers line with the message bytes
    it serializes the stream and headers into a single file
    """

    def serialize(self, message: Message) -> BinaryIO:
        """
        serializes the message into a stream to write to file

        :param message: the message to serialize
        :return: a stream containing the serialized message
        """
        headers_bytes = json.dumps(message.headers, default=json_safe_encoder).encode()
        result = BytesIO()
        result.write(headers_bytes)
        result.write(b'\n')
        result.write(message.bytes)
        result.seek(0)
        return result

    def deserialize(self, stream: BinaryIO) -> Message:
        """
        deserializes the message from the stream

        :param stream: the stream to deserialize from
        :return: the deserialized message
        """
        first_line = stream.readline()
        rest = stream.read()
        headers = json.loads(first_line.decode())
        data = BytesIO(rest)
        return Message(data, headers)


class NoHeadersFileSystemSerializer(FileSystemSerializerBase):
    """
    this is a serializer for filesystem, that ignores the headers, and just puts the stream as the content of the file
    """

    def serialize(self, message: Message) -> BinaryIO:
        """
        serializes the message stream only

        :param message: the message to serialize
        :return: a stream containing the serialized message
        """
        return message.stream

    def deserialize(self, stream: BinaryIO) -> Message:
        """
        deserializes the stream into the stream of the message

        :param stream: the stream to deserialize from
        :return: the deserialized message
        """
        return Message(stream.read(), {})


class ZIPFileSystemSerializer(FileSystemSerializerBase):
    """
    this serializer, creates a zip file, with two inner files: the buffer, and headers
    """
    HEADERS_FILENAME = 'headers'
    BYTES_FILENAME = 'bytes'

    def serialize(self, message: Message) -> BinaryIO:
        """
        serializes the message into a stream to write to file

        :param message: the message to serialize
        :return: a stream containing the serialized message
        """
        headers_bytes = json.dumps(message.headers, default=json_safe_encoder).encode()
        zip_filebuf = BytesIO()

        with zipfile.ZipFile(zip_filebuf, mode='w') as zip_file:
            zip_file.writestr(self.BYTES_FILENAME, message.bytes)
            zip_file.writestr(self.HEADERS_FILENAME, headers_bytes)
        zip_filebuf.seek(0)
        return zip_filebuf

    def deserialize(self, stream: BinaryIO) -> Message:
        """
        deserializes the message from the stream

        :param stream: the stream to deserialize from
        :return: the deserialized message
        """
        with zipfile.ZipFile(stream, mode='r') as zip_file:
            headers_data = zip_file.read(self.HEADERS_FILENAME)
            bytes_data = zip_file.read(self.BYTES_FILENAME)

        headers = json.loads(headers_data.decode())
        data = BytesIO(bytes_data)
        return Message(data, headers)


DefaultFileSystemSerializer = ZIPFileSystemSerializer



from typing import TYPE_CHECKING

try:
    from boto3.session import Config, Session  # type: ignore
except ImportError as ex:
    raise ImportError('Please Install the required extra: messageflux[objectstorage]') from ex

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3ServiceResource

_S3_DEFAULT_TIMEOUT = 1
_S3_DEFAULT_RETRIES = 2


class S3Client:
    """
    This class exists for compatibility purposes only.
    normally, it should not be used
    """

    def __init__(self,
                 endpoint: str,
                 access_key: str,
                 secret_key: str,
                 timeout: int = _S3_DEFAULT_TIMEOUT,
                 retries: int = _S3_DEFAULT_RETRIES):
        """
        represents an s3 client object

        :param endpoint: the endpoint to s3 (url)
        :param access_key: the access key
        :param secret_key: the secret key
        """
        self._s3: 'S3ServiceResource' = Session().resource('s3',
                                                           aws_access_key_id=access_key,
                                                           aws_secret_access_key=secret_key,
                                                           endpoint_url=endpoint,
                                                           verify=False,
                                                           config=Config(
                                                               signature_version='s3',
                                                               connect_timeout=timeout,
                                                               retries={
                                                                   "max_attempts": retries},
                                                               s3={'addressing_style': 'path'}))

    @property
    def s3_resource(self) -> 'S3ServiceResource':
        """
        returns the actual s3 client

        :return: the actual s3 client
        """
        return self._s3



from pathlib import Path

# noinspection PyUnresolvedReferences
from setuptools.build_meta import *

from make_all_extra_requirements import generate_all_requirements

package_dir = Path(__file__).parent.parent
generate_all_requirements(path=package_dir)



import logging
import os
import posixpath
import random
from datetime import datetime

from messageflux.iodevices.base.common import MessageBundle
from messageflux.iodevices.file_system import DefaultFileSystemSerializer
from messageflux.iodevices.message_store_device_wrapper.message_store_base import MessageStoreBase
from messageflux.metadata_headers import MetadataHeaders
from messageflux.utils import get_random_id


class FileSystemMessageStore(MessageStoreBase):
    """
    used as a message store, backed by filesystem
    """

    DATE_FORMAT = "%Y-%m-%d"

    def __init__(self,
                 root_folder: str,
                 num_of_subdirs: int = 4000):
        """
        :param root_folder: the root folders to use
        :param num_of_subdirs: number of subdirectories to create under each date in root_folder
        """
        self._root_folder = root_folder
        self._logger = logging.getLogger(__name__)
        self._num_of_subdirs = num_of_subdirs
        self._serializer = DefaultFileSystemSerializer()

    def connect(self):
        """
        connects to Message Store
        """
        os.makedirs(self._root_folder, exist_ok=True)

    def disconnect(self):
        """
        closes the connection to Message Store
        """
        pass

    @property
    def magic(self) -> bytes:
        """
        return a magic string that is unique and constant for this message store

        """
        return b"__FS_MSGSTORE__"

    def read_message(self, key: str) -> MessageBundle:
        """
        reads a message according to the key given

        :param str key: the key to the message
        :return: the Message
        """
        file_path = self.get_absolute_path(key)
        with open(file_path, 'rb') as f:
            message = self._serializer.deserialize(f)

        return MessageBundle(message, {MetadataHeaders.FILENAME: file_path})

    def generate_relative_path(self) -> str:
        """
        Generate a relative path for the file to be saved as well as loaded from a different machine later.
        Notice the path will always be in posix format regardless of the os in order to be consistent.

        :return: The relative path to use together with the root path in order to save the file.
        """
        filename = get_random_id() + ".FSMS"
        current_date = datetime.now().strftime(self.DATE_FORMAT)
        random_str = str(random.randint(0, self._num_of_subdirs))
        subdir = f'{current_date}-{random_str}'
        return posixpath.join(subdir, filename)

    def get_absolute_path(self, relative_path: str) -> str:
        """
        Convert a relative path that was returned by "generate_relative_path" into the path where the file is at.

        :param relative_path: The key that was returned by "generate_relative_path"
        :return: A path where the original file can be written/read/deleted.
        """
        return os.path.join(self._root_folder, os.path.normpath(relative_path))

    def put_message(self, device_name: str, message_bundle: MessageBundle) -> str:
        """
        puts a message in the message store

        :param str device_name: the name of the device putting the item in the store
        :param message_bundle: the Message bundle to write to the store
        :return: the key to the message in the message store
        """
        # TODO: maybe add the device name to path? need to sanitize it so it's legal dir name
        relative_path = self.generate_relative_path()
        file_path = self.get_absolute_path(relative_path)

        os.makedirs(os.path.dirname(file_path), exist_ok=True)

        with open(file_path, 'wb') as f:
            f.write(self._serializer.serialize(message_bundle.message).read())

        os.chmod(file_path, 0o777)
        return relative_path

    def delete_message(self, key: str):
        """
        deletes a message from the message store

        :param str key: the key to the message
        """
        file_path = self.get_absolute_path(key)
        os.remove(file_path)
        subdir = os.path.dirname(file_path)
        try:
            os.rmdir(subdir)
        except OSError:
            self._logger.warning(f'Could not delete directory {subdir}', exc_info=True)



import os
from pathlib import Path
from typing import List, Set


def load_requirements(path: Path, filename) -> List[str]:
    return (path / filename).read_text().splitlines()


def load_all_extra_requirements(path: Path) -> List[str]:
    import re
    extra_requirement_pattern = re.compile(r'^requirements[_\-](?P<name>.+?)\.txt$')
    files = [f for f in path.iterdir() if f.is_file()]
    all_requirements: Set[str] = set()

    for file in files:
        match = extra_requirement_pattern.match(file.name)
        if match:
            all_requirements.update((path / file.name).read_text().splitlines())

    return sorted([line.strip() for line in all_requirements])


def generate_all_requirements(path: Path, output_filename: str = "requirements-all.txt"):
    all_extras = os.linesep.join(load_all_extra_requirements(path))
    output_file = (path/output_filename)

    print(f'---------------> Writing the requirements to {output_file}:')
    print(all_extras)
    output_file.write_text(all_extras)
    print()
    print('---------------> Done!!!')



import datetime
import os
import threading
from typing import Collection, TypeVar, List, Iterator, Generic, Callable, Optional, Any, Union

from itertools import cycle, islice
from time import perf_counter

EllipsisType = type(...)


class KwargsException(Exception):
    """
    the basic exception type we use.
    adds the ability to add KWARGS to exception
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.kwargs = kwargs


class AggregatedException(KwargsException):
    """
    an exception that can hold multiple inner exceptions
    """

    def __init__(self, *args, inner_exceptions: Optional[List[Exception]] = None, **kwargs):
        super(AggregatedException, self).__init__(*args, **kwargs)
        self.inner_exceptions = inner_exceptions


TItemType = TypeVar('TItemType')


class StatefulListIterator(Collection[TItemType]):
    """
    An list, that can be iterated from the last stopped location, in a circular fashion.
    every iteration starts from the last position iterated position, and iterates over all the list
    """

    def __init__(self, items: List[TItemType]) -> None:
        self._items = items
        self._item_count = len(items)
        self._item_cycle = cycle(self._items)

    def __contains__(self, __x: object) -> bool:
        return self._items.__contains__(__x)

    def __len__(self) -> int:
        return self._item_count

    def __iter__(self) -> Iterator[TItemType]:
        return islice(self._item_cycle, self._item_count)


TEventType = TypeVar('TEventType')


class ObservableEvent(Generic[TEventType]):
    """
    an event that can be subscribed to and fired.
    """

    def __init__(self) -> None:
        self._handlers: List[Callable[[TEventType], None]] = []

    def subscribe(self, handler: Callable[[TEventType], None]) -> None:
        """
        subscribes a callback to the event

        :param handler: the callback to subscribe
        """
        self._handlers.append(handler)

    def unsubscribe(self, handler: Callable[[TEventType], None]) -> bool:
        """
        unsubscribes a callback to the event

        :param handler: the callback to unsubscribe
        :return: True if the callback existed and unsubscribed. False if the callback didn't exist
        """
        if handler in self._handlers:
            self._handlers.remove(handler)
            return True
        return False

    def fire(self, event: TEventType, continue_after_failure: bool = True) -> bool:
        """
        fires the event

        :param event: the event to fire
        :param continue_after_failure: True means to call all the callbacks even if one fails.
        False means to stop if one of the callbacks fails
        :return: True if all the callbacks succeeded, False otherwise
        """
        ret_val = True
        for handler in self._handlers:
            try:
                handler(event)
            except Exception:
                ret_val = False
                if not continue_after_failure:
                    raise
        return ret_val


TValueType = TypeVar('TValueType')


class ThreadLocalValue(threading.local, Generic[TValueType]):
    """
    this class holds a single, thread-local value

    x = ThreadLocalValue(5)
    x.value # 5
    x.value = 7 # changes only for this thread
    """

    def __init__(self, init_value: Optional[TValueType] = None):
        self.value = init_value


class ThreadLocalMember(Generic[TValueType]):
    """
    this is a descriptor, for making thread local members for class:

    class MyClass:
        x = ThreadLocalMember()
        y = ThreadLocalMember(1)
        z = ThreadLocalMember()
        w = ThreadLocalMember(1)
        def __init__():
            self.x = 5
            self.w = 2

    a = MyClass():
    a.x # 5
    a.x = 7 # changes only for this thread
    a.y # 1 - the default init value for this
    a.z # raises AttributeError (z was not set on class). once it's set, then value is the init value for each thread
    a.w # 2 for this thread, 1 for any other thread (because of the default init value)
    """

    def __init__(self, init_value: Union[TValueType, EllipsisType] = ...):  # type: ignore
        self._init_value = init_value

    def __set_name__(self, owner, name):
        self._public_name = name
        self._private_name = f'_thread_local_{name}'

    def _get_thread_local_value(self, instance, init_value):
        try:
            lv = getattr(instance, self._private_name)
        except AttributeError:
            lv = ThreadLocalValue(init_value)
            setattr(instance, self._private_name, lv)
        return lv

    def __get__(self, instance, owner) -> Optional[TValueType]:
        if self._init_value is not ...:
            lv = self._get_thread_local_value(instance, self._init_value)
        else:
            if not hasattr(instance, self._private_name):
                raise AttributeError(f"'{owner.__name__}' object has no attribute '{self._public_name}'")
            lv = getattr(instance, self._private_name)

        return lv.value

    def __set__(self, instance, value):
        if instance is None:
            return
        if self._init_value is ...:
            init_value = value
        else:
            init_value = self._init_value
        lv = self._get_thread_local_value(instance, init_value)
        lv.value = value


def get_random_id() -> str:
    """
    generates a random 16 bytes hex string
    """
    return os.urandom(16).hex()


class ContextDidNotEndException(KwargsException):
    """
    this exception is raised when elapsed seconds is called on a context that was not run
    """
    pass


class TimerContext:
    """
    a context object that records how much time did the context last
    """

    def __init__(self):
        self._start: float = perf_counter()
        self._elapsed: Optional[float] = None

    @property
    def elapsed_seconds(self) -> float:
        """
        returns the elapsed time for this context (between __enter__ and __exit__)

        :return: the elapsed time (in seconds) for the context.
        if __enter__ or __exit__ weren't called, a ContextDidNotEndException exception is raised
        """
        if self._elapsed is None:
            raise ContextDidNotEndException("Context did not end")
        return self._elapsed

    def __enter__(self):
        self._start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._elapsed = perf_counter() - self._start


def json_safe_encoder(obj: Any):
    """
    Convert non-serializable types to strings.

    :param obj: The object to make serializable.
    :return: Serializable version of the object (string-form).
    """

    try:
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.time):
            return obj.strftime('%H:%M')
        elif isinstance(obj, bytes):
            return obj.decode()
        else:
            return str(obj)
    except Exception:
        return "UNENCODABLE"



import json
import re
from datetime import datetime
from io import BytesIO
from typing import Optional, Dict, Any, Iterator, TYPE_CHECKING, Union, IO
from typing_extensions import Literal
from urllib.parse import urljoin

try:
    from boto3.s3.inject import ClientError
except ImportError as ex:
    raise ImportError('Please Install the required extra: messageflux[objectstorage]') from ex

BUCKET_NAME_VALIDATOR = re.compile(r'^[a-z0-9][a-z0-9.\-]{1,61}[a-z0-9]$')

if TYPE_CHECKING:
    from mypy_boto3_s3.type_defs import LifecycleConfigurationBucketLifecycleTypeDef, GetObjectOutputTypeDef
    from mypy_boto3_s3 import S3ServiceResource
    from mypy_boto3_s3.service_resource import ObjectSummary
    from _typeshed import SupportsRead


class S3BucketException(Exception):
    """
    represents errors thrown from S3 bucket
    """
    pass


class S3NoSuchItem(S3BucketException):
    """
    represents no such item in bucket
    """
    pass


class S3NoSuchBucket(S3BucketException):
    """
    represents no such bucket
    """
    pass


class S3Object:
    """
    represents an s3 object with lazy content get
    """

    def __init__(self, object_summery: 'ObjectSummary', object_cache: Optional['GetObjectOutputTypeDef'] = None):
        self._object_summery = object_summery
        self._object_cache = object_cache
        self._body: Optional[BytesIO] = None
        self._metadata: Optional[Dict[str, str]] = None

    @property
    def _object_dict(self) -> 'GetObjectOutputTypeDef':
        if self._object_cache is None:
            self._object_cache = self._object_summery.get()
        return self._object_cache

    @property
    def key(self) -> str:
        """
        the object's key
        """
        return self._object_summery.key

    @property
    def body_stream(self) -> 'SupportsRead[bytes]':
        """
        the stream for the object's body (lazy read)
        """
        if self._body is None:
            return self._object_dict['Body']
        else:
            return self._body

    @property
    def body(self) -> BytesIO:
        """
        the bytes for the object's body (lazy read)
        """
        if self._body is None:
            self._body = BytesIO(self._object_dict['Body'].read())
        return self._body

    @property
    def metadata(self) -> Dict[str, str]:
        """
        the object's metadata
        """
        if self._metadata is None:
            self._metadata = self._object_dict.get('Metadata', {})

        assert self._metadata is not None
        return self._metadata

    @property
    def last_modified(self) -> datetime:
        """
        the object's last modified time (might incur a request to the server)
        """
        return self._object_summery.last_modified

    @property
    def size(self) -> int:
        """
        the object's size (might incur a request to the server)
        """
        return self._object_summery.size


class S3Bucket:
    """
    this class represents a single S3 Bucket
    """

    @staticmethod
    def create_bucket(s3_resource: 'S3ServiceResource',
                      bucket_name: str,
                      lifetime_in_days: Optional[int] = None,
                      allow_public_access=False) -> 'S3Bucket':
        """
        creates and returns a bucket

        :param s3_resource: the s3 resource to use
        :param bucket_name: the bucket name to create and get
        :param lifetime_in_days: the lifetime of objects in the bucket to set (None for no change)
        :param allow_public_access: should we make the bucket publicly accessible from web
        :return: the bucket
        """
        bucket = S3Bucket(bucket_name=bucket_name, s3_resource=s3_resource, auto_create=True)
        if lifetime_in_days is not None:
            bucket.set_lifetime_in_days(lifetime_in_days)

        if allow_public_access:
            bucket.allow_public_access()

        return bucket

    @staticmethod
    def list_buckets(s3_resource: 'S3ServiceResource') -> Iterator['S3Bucket']:
        """
        returns a list of all the buckets in this client

        :param s3_resource: the s3 resource to use
        :return: iterator of bucket objects
        """
        for bucket in s3_resource.buckets.all():
            yield S3Bucket(bucket_name=bucket.name, s3_resource=s3_resource)

    def __init__(self, bucket_name: str, s3_resource: 'S3ServiceResource', auto_create: bool = False):
        """
        this class represents a single S3 Bucket

        :param bucket_name: the name of the bucket to work with
        :param s3_resource: the s3 resource to use
        :param auto_create: should we create the bucket if it doesn't exist?
        """
        if not BUCKET_NAME_VALIDATOR.match(bucket_name):
            raise S3BucketException(f'Invalid Bucket Name:{bucket_name}. check bucket naming rules.')
        try:
            self._bucket_name = bucket_name
            self._s3_resource = s3_resource
            self._s3bucket = s3_resource.Bucket(bucket_name)
            s3_resource.meta.client.head_bucket(Bucket=bucket_name)
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')

            if code in ['NoSuchKey', '404']:
                if auto_create:
                    self._s3bucket.create()
                else:
                    raise S3NoSuchBucket(f'No bucket with name {bucket_name} exists')
            else:
                raise S3BucketException(
                    f'Error While creating S3Bucket "{bucket_name}": {code}') from ex

    def delete_bucket(self, force=False):
        """
        deletes all items in bucket, and then deletes bucket
        :param force: True will delete all objects on bucket, before deleting the bucket
        """
        if force:
            self.clear_objects()
        self._s3bucket.delete()

    @property
    def s3_resource(self) -> 'S3ServiceResource':
        """
        the client for this bucket
        """
        return self._s3_resource

    @property
    def name(self) -> str:
        """
        returns the bucket name
        :return: the bucket name
        """
        return self._bucket_name

    def compute_url(self, key):
        """
        computes the item url, by key

        :param key: the key to item
        :return: the url to item
        """
        return urljoin(self.s3_resource.meta.client.meta.endpoint_url, f'{self.name}/{key}')

    def allow_public_access(self):
        """
        makes the bucket publicly accessible from web
        """
        policy = {
            "Statement": [
                {
                    "Effect": "Allow",
                    "Principal": "*",
                    "Action": [
                        "s3:GetObject",
                        "s3:ListBucket"
                    ],
                    "Resource": [
                        f"arn:aws:s3:::{self.name}",
                        f"arn:aws:s3:::{self.name}/*"
                    ]
                }
            ]
        }
        self._s3bucket.Policy().put(Policy=json.dumps(policy))

    def set_lifetime_in_days(self, days: int):
        """
        sets the lifetime of all objects in bucket to 'days' days (or <=0 to cancel expiration)

        :param days: the number of days for an item to live
        """
        try:
            status: Literal['Disabled', 'Enabled']
            if days <= 0:
                status = 'Disabled'
                days = 1
            else:
                status = 'Enabled'
            lc: 'LifecycleConfigurationBucketLifecycleTypeDef' = {
                'Rules': [{
                    'Status': status,
                    'Prefix': '',
                    'Expiration': {
                        'Days': days
                    },
                    'ID': 'expire'
                }]
            }
            self._s3bucket.Lifecycle().put(LifecycleConfiguration=lc)
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')
            raise S3BucketException(
                f'Error While setting bucket lifecycle: {code}') from ex

    def put_object(self, key: str, buf: Union[str, bytes, IO[Any]], metadata: Dict[str, str], **kwargs) -> S3Object:
        """
        puts a binary object in the bucket

        :param key: the key of the object to put
        :param buf: the buffer to put in the bucket
        :param metadata: extra metadata
        """
        try:
            self._s3bucket.put_object(Key=key, Metadata=metadata, Body=buf, **kwargs)
            result = S3Object(object_summery=self.s3_resource.ObjectSummary(bucket_name=self._bucket_name, key=key))
            return result
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')
            raise S3BucketException(
                f'Error While putting object to key "{key}": {code}') from ex

    def upload_object(self, key: str, stream: IO[Any], metadata: Dict[str, str], **kwargs) -> S3Object:
        """
        uploads a binary stream to the bucket

        :param key: the key of the object to put
        :param stream: the stream to put in the bucket
        :param metadata: extra metadata
        """
        try:
            extra_args = kwargs.pop('ExtraArgs', {})
            extra_args['Metadata'] = metadata
            self._s3_resource.meta.client.upload_fileobj(Fileobj=stream,
                                                         Bucket=self.name,
                                                         Key=key,
                                                         ExtraArgs=extra_args,
                                                         **kwargs)
            return S3Object(self.s3_resource.ObjectSummary(bucket_name=self._bucket_name, key=key))
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')
            raise S3BucketException(
                f'Error While uploading object to key "{key}": {code}') from ex

    def download_object(self, key: str, writable_stream: IO[Any], **kwargs):
        """
        downloads an object from the bucket into a writable stream

        :param key: the key of the object to download
        :param writable_stream: the writable stream to write into
        :return:
        """
        self.s3_resource.meta.client.download_fileobj(self.name, key, writable_stream, **kwargs)

    def get_object(self, key: str) -> S3Object:
        """
        gets a binary from the bucket

        :param key: the key of the binary
        :return: the buffer of the object, and the metadata dict
        """
        try:
            obj_summary = self.s3_resource.ObjectSummary(bucket_name=self._bucket_name, key=key)
            return S3Object(obj_summary, obj_summary.get())
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')

            if code in ['NoSuchKey', '404']:
                raise S3NoSuchItem(f'No item with key {key} in bucket {self._bucket_name}')
            else:
                raise S3BucketException(
                    f'Error While getting object from key "{key}": {code}') from ex

    def delete_object(self, key: str):
        """
        deletes a binary from the bucket

        :param key: the key of the binary
        """
        try:
            self._s3bucket.Object(key).delete()
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')
            if code in ['NoSuchKey', '404']:
                raise S3NoSuchItem(f'No item with key {key} in bucket {self._bucket_name}')
            else:
                raise S3BucketException(
                    f'Error While deleting object in key "{key}": {code}') from ex

    def list_objects(self) -> Iterator[S3Object]:
        """
        lists all the objects in this bucket
        :return: a generator of S3Objects
        """
        try:
            response = self._s3bucket.objects.all()
            for item in response:
                yield S3Object(item)
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')
            raise S3BucketException(
                f'Error While listing objects in bucket "{self._bucket_name}": {code}') from ex

    def find_objects(self,
                     prefix: Optional[str] = None,
                     delimiter: Optional[str] = None,
                     max_keys: int = 1000,
                     **kwargs) -> Iterator[S3Object]:
        """
        find objects ih the bucket.

        :param delimiter: A Delimiter is a character you use to group keys
        :param prefix: Limits the response to keys that begin with the specified prefix
        :param max_keys: Sets the maximum number of keys returned in the response
        :param **kwargs: additional filters to s3
        :return: dict from object key to last modified
        """
        try:
            if delimiter is not None:
                kwargs['Delimiter'] = delimiter
            if prefix is not None:
                kwargs['Prefix'] = prefix

            response = self._s3bucket.objects.filter(MaxKeys=max_keys,
                                                     **kwargs)
            for item in response:
                yield S3Object(item)
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')
            raise S3BucketException(
                f'Error While finding objects in bucket "{self._bucket_name}": {code}') from ex

    def clear_objects(self):
        """
        clears all the objects in the bucket
        """
        try:
            self._s3bucket.objects.all().delete()
        except ClientError as ex:
            code = ''
            if 'Error' in ex.response:
                code = ex.response['Error'].get('Code', '')
            raise S3BucketException(
                f'Error While removing all objects in bucket "{self._bucket_name}": {code}') from ex



from .file_system_input_device import (FileSystemInputDeviceManager,
                                       FileSystemInputDevice,
                                       FileSystemInputTransaction)

from .file_system_output_device import (FileSystemOutputDeviceManager,
                                        FileSystemOutputDevice)

from .file_system_serializer import (ZIPFileSystemSerializer,
                                     DefaultFileSystemSerializer,
                                     NoHeadersFileSystemSerializer)



import logging

from messageflux.iodevices.base import OutputDevice, OutputDeviceException, OutputDeviceManager
from messageflux.iodevices.base.common import MessageBundle
from typing import Optional


class FailoverOutputDevice(OutputDevice['FailoverOutputDeviceManager']):
    """
    this device tries to send to primary output device - if the primary fails, it sends to secondary
    """

    def __init__(self,
                 device_manager: 'FailoverOutputDeviceManager',
                 inner_device: OutputDevice,
                 failover_device: OutputDevice,
                 device_name: Optional[str] = None):
        """
        this device tries to send to primary output device - if the primary fails, it sends to secondary

        :param inner_device: the device to wrap
        :param failover_device: the failover device to send to in case of primary fail
        :param device_name: the name for the device (or None to take from inner_device
        """
        if device_name is None:
            device_name = inner_device.name
        super(FailoverOutputDevice, self).__init__(device_manager, device_name)
        self._inner_device = inner_device
        self._failover_device = failover_device
        self._logger = logging.getLogger(__name__)

    @property
    def inner_device(self) -> OutputDevice:
        """
        the inner device
        :return: the inner device
        """
        return self._inner_device

    @property
    def failover_device(self) -> OutputDevice:
        """
        the failover device
        :return: the failover device
        """
        return self._failover_device

    def close(self):
        """
        closes the underlying devices
        """
        primary_worked = False
        try:
            self._inner_device.close()
            primary_worked = True
        except Exception:
            self._logger.warning(f'Error disconnecting from underlying device {self._inner_device.name}', exc_info=True)

        try:
            self._failover_device.close()
        except Exception:
            if primary_worked:
                self._logger.warning(f'Error disconnecting from FAILOVER underlying device {self._inner_device.name}',
                                     exc_info=True)
            else:
                raise

    def _send_message(self, message_bundle: MessageBundle):
        """
        does the actual sending of a stream to OutputDevice with metadata

        :param stream: the stream to send
        """
        failures = []
        try:
            self._inner_device.send_message(message_bundle.message, message_bundle.device_headers)
            return
        except Exception as ex:
            failures.append(ex)
            self._logger.warning(f'Error sending to underlying device {self._inner_device.name}', exc_info=True)

            try:
                self._failover_device.send_message(message_bundle.message.copy(), message_bundle.device_headers)
                return
            except Exception as ex:
                failures.append(ex)
                raise OutputDeviceException("Couldn't send to device or failover device", inner_exceptions=failures)


class FailoverOutputDeviceManager(OutputDeviceManager[OutputDevice]):
    """
    Output Device Manager that uses a failover if the send fails
    """

    def __init__(self, inner_device_manager: OutputDeviceManager, failover_device_manager: OutputDeviceManager):
        """
        :param inner_device_manager: the device to wrap
        :param failover_device_manager: the failover device_manager to send to in case of primary fail
        """
        self._inner_device_manager = inner_device_manager
        self._failover_device_manager = failover_device_manager
        self._logger = logging.getLogger(__name__)

    def connect(self):
        """
        connects to device manager
        """
        primary_worked = False
        try:
            self._inner_device_manager.connect()
            primary_worked = True
        except Exception:
            self._logger.warning(
                f'Error connecting to underlying manager: {type(self._inner_device_manager).__name__}', exc_info=True)
        try:
            self._failover_device_manager.connect()
        except Exception:
            if primary_worked:
                self._logger.warning(
                    f'Error connecting to underlying manager: {type(self._failover_device_manager).__name__}',
                    exc_info=True)
            else:
                raise

    def disconnect(self):
        """
        closes the connection to IODeviceManager
        """
        primary_worked = False
        try:
            self._inner_device_manager.disconnect()
            primary_worked = True
        except Exception:
            self._logger.warning(
                f'Error disconnecting from underlying manager: {type(self._inner_device_manager).__name__}',
                exc_info=True)
        try:
            self._failover_device_manager.disconnect()
        except Exception:
            if primary_worked:
                self._logger.warning(
                    f'Error disconnecting from underlying manager: {type(self._failover_device_manager).__name__}',
                    exc_info=True)
            else:
                raise

    def get_output_device(self, name: str) -> OutputDevice:
        """
        Returns an output device by name

        :param name: the name of the device to write to
        :return: an output device for 'device_name'
        """
        inner_device = None
        try:
            inner_device = self._inner_device_manager.get_output_device(name)
        except Exception:
            self._logger.warning(
                f'Error creating output device {name} from {type(self._inner_device_manager).__name__} manager',
                exc_info=True)
        try:
            failover_device = self._failover_device_manager.get_output_device(name)
        except Exception:
            if inner_device is not None:
                self._logger.warning(
                    f'Error creating output device {name} from {type(self._failover_device_manager).__name__} '
                    f'manager',
                    exc_info=True)
                return inner_device
            else:
                raise

        if inner_device is None:
            return failover_device

        return FailoverOutputDevice(device_manager=self,
                                    inner_device=inner_device,
                                    failover_device=failover_device,
                                    device_name=name)



import logging
from typing import BinaryIO

from messageflux.iodevices.message_store_device_wrapper.message_store_base import MessageStoreBase


class _MessageStoreTransformerBase:
    """
    this is a base class for message store transformer
    """
    MAGIC_HEADER = b"__MSGSTORE_WRAPPER__"

    def __init__(self, message_store: MessageStoreBase):
        """
        :param message_store: the message store to use
        """
        self._message_store = message_store
        self._logger = logging.getLogger(__name__)
        self._full_magic = b"|".join([self.MAGIC_HEADER,
                                      self._message_store.magic
                                      ])

    def __enter__(self) -> '_MessageStoreTransformerBase':
        self._message_store.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._message_store.__exit__(exc_type, exc_val, exc_tb)

    def connect(self):
        """
        connects to device message store
        """
        self._message_store.connect()

    def disconnect(self):
        """
        disconnects from the message store
        """
        self._message_store.disconnect()

    def serialize_key(self, key: str) -> bytes:
        """
        serializes the key for sending on the wire

        :param str key: the key
        :return: serialized key to send
        """
        return self._full_magic + key.encode()

    def deserialize_key(self, data: bytes) -> str:
        """
        deserializes the key received from the wire

        :param bytes data: the data from the wire
        :return: deserialized key
        """
        key = data[len(self._full_magic):]
        return key.decode()

    def is_key(self, stream: BinaryIO) -> bool:
        """
        checks if the data from the wire, is a serialized key

        :param stream: the data from the wire
        :return: True if it is a serialized key
        """
        magic = stream.read(len(self._full_magic))
        stream.seek(0)
        return magic == self._full_magic



import zlib

from messageflux.iodevices.base import Message, ReadResult
from messageflux.iodevices.base.common import MessageBundle
from messageflux.iodevices.transformer_device_wrapper import InputTransformerBase, OutputTransformerBase
from messageflux.iodevices.transformer_device_wrapper.transformer_input_device import TransformerInputDevice
from messageflux.iodevices.transformer_device_wrapper.transformer_output_device import TransformerOutputDevice


class ZLIBTransformer(InputTransformerBase, OutputTransformerBase):
    """
    This class uses zlib to compress data before sending it to the underlying device, and decompress data coming out of
    the underlying device, if necessary.
    """
    ZLIB_TRANSFORMER_MAGIC = b'__ZLIBTRANSFORMER__'

    def __init__(self, level: int = -1):
        """

        :param level: the compression level for zlib. only matters for outgoing messages
        """
        self._level = level
        if not (-1 <= self._level <= 9):
            raise ValueError("ZLIBTransformer: level must be between -1 and 9")

    def transform_outgoing_message(self, output_device: TransformerOutputDevice,
                                   message_bundle: MessageBundle) -> MessageBundle:

        compressed_data = self.ZLIB_TRANSFORMER_MAGIC + zlib.compress(message_bundle.message.bytes, level=self._level)
        return MessageBundle(message=Message(compressed_data, message_bundle.message.headers),
                             device_headers=message_bundle.device_headers)

    def transform_incoming_message(self, input_device: TransformerInputDevice, read_result: ReadResult) -> ReadResult:
        if read_result.message.stream.read(len(self.ZLIB_TRANSFORMER_MAGIC)) != self.ZLIB_TRANSFORMER_MAGIC:
            # This is not a zlib-compressed message
            read_result.message.stream.seek(0)
            return read_result

        decompressed_data = zlib.decompress(read_result.message.stream.read())
        return ReadResult(message=Message(decompressed_data, read_result.message.headers),
                          device_headers=read_result.device_headers,
                          transaction=read_result.transaction)



from abc import ABCMeta, abstractmethod
from typing import Optional, TypeVar, Generic

from messageflux.iodevices.base.common import Message, DeviceHeaders, MessageBundle
from messageflux.utils import AggregatedException

TManagerType = TypeVar('TManagerType', bound='OutputDeviceManager')
TOutputDeviceType = TypeVar('TOutputDeviceType', bound='OutputDevice')


class OutputDeviceException(AggregatedException):
    """
    a base exception class for all output device related exceptions
    """
    pass


class OutputDevice(Generic[TManagerType], metaclass=ABCMeta):
    """
    base class for all output devices
    """

    def __init__(self, manager: TManagerType, name: str):
        """

        :param manager: the output device manager that created this device
        :param name: the name of this device
        """
        self._manager = manager
        self._name = name

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def name(self) -> str:
        """
        :return: the name of this device
        """
        return self._name

    @property
    def manager(self) -> TManagerType:
        """
        :return: the input device manager that created this device
        """
        return self._manager

    def send_message(self, message: Message, device_headers: Optional[DeviceHeaders] = None):
        """
        sends a message to the device.

        :param message: the message to send
        :param device_headers: optional headers to send to underlying device.
        those headers are not part of the message, but contains extra data for the device, that can modify its operation
        """
        device_headers = device_headers or {}
        self._send_message(MessageBundle(message=message, device_headers=device_headers))

    @abstractmethod
    def _send_message(self, message_bundle: MessageBundle):
        """
        sends a message to the device. this should be implemented by child classes

        :param message_bundle: the message bundle to send
        """
        pass

    def close(self):
        """
        and optional method that cleans device resources if necessary
        """
        pass


class OutputDeviceManager(Generic[TOutputDeviceType], metaclass=ABCMeta):
    """
    this is a base class for output device managers. it is used to create output devices
    """

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self):
        """
        connects to the device manager
        """
        pass

    def disconnect(self):
        """
        disconnects from the device manager
        """
        pass

    @abstractmethod
    def get_output_device(self, name: str) -> TOutputDeviceType:
        """
        creates an output device. this should be implemented by child classes

        :param name: the name of the output device to create
        :return: the created output device
        """
        pass



import logging
import ssl
from typing import BinaryIO, Dict, Any, Union, Optional, List, TYPE_CHECKING

import time

from messageflux.iodevices.base import OutputDevice, OutputDeviceException, OutputDeviceManager
from messageflux.iodevices.base.common import MessageBundle
from messageflux.iodevices.rabbitmq.rabbitmq_device_manager_base import RabbitMQDeviceManagerBase
from messageflux.metadata_headers import MetadataHeaders
from messageflux.utils import get_random_id, ThreadLocalMember

if TYPE_CHECKING:
    from pika.adapters.blocking_connection import BlockingChannel


class RabbitMQOutputDevice(OutputDevice['RabbitMQOutputDeviceManager']):
    """
    represents an RabbitMQ output devices
    """

    def __init__(self, device_manager: 'RabbitMQOutputDeviceManager', routing_key: str, exchange: str = ''):
        """
        constructs a new output RabbitMQ device

        :param device_manager: the RabbitMQ device Manager that holds this device
        :param routing_key: the routing key for this queue
        :param exchange: the exchange name in RabbitMQ for this output device
        """
        super(RabbitMQOutputDevice, self).__init__(device_manager, routing_key)
        self._routing_key = routing_key
        self._exchange = exchange
        self._logger = logging.getLogger(__name__)

    def _send_message(self, message_bundle: MessageBundle):
        rabbit_headers = self.manager.default_rabbit_headers.copy()
        custom_headers = message_bundle.device_headers
        # if specific rabbitmq headers we want them to override other headers
        rabbit_headers.update(custom_headers)
        self.manager.publish_message(
            routing_key=self._routing_key,
            data=message_bundle.message.stream,
            exchange=self._exchange,
            headers=message_bundle.message.headers,
            app_id=rabbit_headers.get('app_id', "BASESERVICE"),
            message_id=rabbit_headers.get('message_id',
                                          message_bundle.message.headers.get(MetadataHeaders.ITEM_ID, get_random_id())),
            persistent=rabbit_headers.get('persistent', True),
            mandatory=rabbit_headers.get('mandatory', self.manager.publish_confirm),
            priority=rabbit_headers.get('priority', None),
            expiration=rabbit_headers.get('expiration', None)
        )


class LengthValidationException(OutputDeviceException):
    """
    this exception is used when a length validation error occurs in IODevice
    """
    pass


class RabbitMQOutputDeviceManager(RabbitMQDeviceManagerBase, OutputDeviceManager[RabbitMQOutputDevice]):
    """
    this manager is used to create RabbitMQ devices
    """
    _PUBLISH_CONFIRM_HEADER = "__RABBITMQ_PUBLISH_CONFIRM__"

    _outgoing_channel: Union[ThreadLocalMember[Optional['BlockingChannel']],
                             Optional['BlockingChannel']] = ThreadLocalMember(init_value=None)

    def __init__(self,
                 hosts: Union[List[str], str],
                 user: str,
                 password: str,
                 port: Optional[int] = None,
                 ssl_context: Optional[ssl.SSLContext] = None,
                 virtual_host: Optional[str] = None,
                 default_output_exchange: str = '',
                 publish_confirm: bool = True,
                 client_args: Optional[Dict[str, str]] = None,
                 heartbeat: int = 300,
                 connection_attempts: int = 5,
                 max_message_length: int = -1,
                 max_header_value_length: int = 2048,
                 max_header_name_length: int = 1024,
                 default_rabbit_headers: Optional[Dict[str, Any]] = None,
                 blocked_connection_timeout: Optional[float] = None,
                 ):
        """
        This manager used to create RabbitMQ devices (direct queues)

        :param hosts: the hostname or a list of hostnames of the manager
        :param user: the username for the rabbitMQ manager
        :param password: the password for the rabbitMQ manager
        :param port: the port to connect the hosts to
        :param ssl_context: the ssl context to use. None means don't use ssl at all
        :param virtual_host: the virtual host to connect to
        :param default_output_exchange: the default exchange used for output devices
        :param publish_confirm: should the send fail if the message is unroutable?
        :param client_args: the arguments to create the client with
        :param int heartbeat: heartbeat interval for the connection (between 0 and 65536
        :param int connection_attempts: Maximum number of retry attempts
        :param int max_message_length: the maximum size (in bytes) of a message to send (0 and below = no limit)
        :param int max_header_value_length: the maximum length of a header's value (0 and below = no limit)
        :param int max_header_name_length: the maximum length of a header's name (0 and below = no limit)
        :param dict default_rabbit_headers: The default headers to use when sending
        :param blocked_connection_timeout: If not None,
            the value is a non-negative timeout, in seconds, for the
            connection to remain blocked (triggered by Connection.Blocked from
            broker); if the timeout expires before connection becomes unblocked,
            the connection will be torn down, triggering the adapter-specific
            mechanism for informing client app about the closed connection:
            passing `ConnectionBlockedTimeout` exception to on_close_callback
            in asynchronous adapters or raising it in `BlockingConnection`.
        """
        super(RabbitMQOutputDeviceManager, self).__init__(hosts=hosts,
                                                          user=user,
                                                          password=password,
                                                          port=port,
                                                          ssl_context=ssl_context,
                                                          virtual_host=virtual_host,
                                                          client_args=client_args,
                                                          connection_type="Output",
                                                          heartbeat=heartbeat,
                                                          connection_attempts=connection_attempts,
                                                          blocked_connection_timeout=blocked_connection_timeout)

        self._default_output_exchange = default_output_exchange
        self._publish_confirm = publish_confirm
        self._max_message_length = max_message_length
        self._max_header_value_length = max_header_value_length
        self._max_header_name_length = max_header_name_length
        self._default_rabbit_headers = {
            'persistent': True,
            'mandatory': self.publish_confirm,
            'priority': None,
            'expiration': None
        }

        if default_rabbit_headers is None:
            default_rabbit_headers = {}

        self._default_rabbit_headers.update(default_rabbit_headers)

    @property
    def default_rabbit_headers(self) -> Dict[str, Any]:
        """
        returns the default rabbit headers
        """
        return self._default_rabbit_headers

    @property
    def publish_confirm(self) -> bool:
        """
        returns the publish_confirm flag
        """
        return self._publish_confirm

    def publish_message(self,
                        routing_key: str,
                        data: BinaryIO,
                        exchange: str = "",
                        headers: Optional[Dict[str, Any]] = None,
                        app_id: Optional[str] = None,
                        message_id: Optional[str] = None,
                        persistent: bool = True,
                        mandatory: bool = False,
                        priority: Optional[int] = None,
                        expiration: Optional[int] = None):
        """
        publishes a message, and reconnects if there's an error

        :param routing_key: The routing key to bind on
        :param data: the stream to send
        :param exchange: The exchange to publish to
        :param headers: the headers to send to queue
        :param str app_id: the app_id to send with the message
        :param str message_id: the id to five the message
        :param bool persistent: should the message be persisted to disk
        :param bool mandatory: The mandatory flag
        :param int priority: the message priority (0-9)
        :param expiration: ttl (in ms) of message
        """
        try:
            from pika.exceptions import (AMQPConnectionError,
                                         UnroutableError,
                                         AMQPChannelError,
                                         ConnectionBlockedTimeout,
                                         NackError)
        except ImportError as ex:
            raise ImportError('Please Install the required extra: messageflux[rabbitmq]') from ex

        if headers is not None:
            # Safety measure for removing input device headers
            for header_name, header_value in headers.items():
                if hasattr(header_name, '__len__') and len(header_name) > self._max_header_name_length > 0:
                    raise LengthValidationException(
                        f"Header name ({header_name}) is longer than {self._max_header_name_length}")
                if hasattr(header_value, '__len__') and len(header_value) > self._max_header_value_length > 0:
                    raise LengthValidationException(
                        f"Value of header {header_name} is longer than {self._max_header_value_length} "
                        f"(header value is {header_value})")

        length = data.seek(0, 2)
        if length >= self._max_message_length > 0:
            raise LengthValidationException(f"file is too big for rabbitmq: {routing_key}, length: {length}")
        data.seek(0)

        try:
            self._inner_publish(routing_key, data, exchange, headers, app_id, message_id,
                                persistent, mandatory, priority, expiration)
        except NackError:
            self._logger.exception(
                f'Publish on routing key "{routing_key}" was NACKed. (possibly the queue reached its max size)')
            raise
        except UnroutableError:
            self._logger.exception(f"could not route message. routing key: {routing_key}")
            raise
        except ConnectionBlockedTimeout:
            self._logger.exception(
                f"could not publish message, due to ConnectionBlocked (possibly Memory Alarm). "
                f"routing key: {routing_key}")
            raise

        except (AMQPConnectionError, AMQPChannelError):
            self._logger.warning(f"failed to send message to queue. routing key: {routing_key}", exc_info=True)
            try:
                self._inner_publish(routing_key, data, exchange, headers, app_id, message_id,
                                    persistent, mandatory, priority, expiration)
            except (AMQPConnectionError, AMQPChannelError):
                self._logger.exception(f"failed to send message to queue. routing key: {routing_key}")
                raise

        except Exception as ex:
            raise OutputDeviceException("Error while sending to RabbitMQ") from ex

    def _inner_publish(self,
                       routing_key: str,
                       data: BinaryIO,
                       exchange: str = "",
                       headers: Optional[Dict[str, Any]] = None,
                       app_id: Optional[str] = None,
                       message_id: Optional[str] = None,
                       persistent: bool = True,
                       mandatory: bool = False,
                       priority: Optional[int] = None,
                       expiration: Optional[int] = None):
        """
        does the actual publish to queue

        :param routing_key: The routing key to bind on
        :param data: the stream to send
        :param exchange: The exchange to publish to
        :param headers: the headers to send to queue
        :param str app_id: the app_id to send with the message
        :param str message_id: the id to five the message
        :param bool persistent: should the message be persisted to disk
        :param bool mandatory: The mandatory flag
        :param int priority: the message priority (0-9)
        :param int expiration: ttl (in ms) of message
        """
        try:
            import pika
        except ImportError as ex:
            raise ImportError('Please Install the required extra: messageflux[rabbitmq]') from ex

        if headers is None:
            headers = {}
        else:
            headers = headers.copy()
        headers[self._PUBLISH_CONFIRM_HEADER] = self.publish_confirm

        data.seek(0)

        str_expiration: Optional[str] = None
        if expiration is not None:
            str_expiration = str(expiration)

        properties = pika.BasicProperties(delivery_mode=2 if persistent else 1,
                                          headers=headers,
                                          app_id=app_id,
                                          message_id=message_id,
                                          timestamp=int(time.time()),
                                          priority=priority,
                                          expiration=str_expiration)

        channel = self.get_outgoing_channel()
        channel.basic_publish(exchange, routing_key, data.read(), properties,
                              mandatory=mandatory)

    def get_outgoing_channel(self) -> 'BlockingChannel':
        """
        returns a rabbitmq channel for publishing messages

        :return: a rabbit mq channel
        """
        try:
            if self._outgoing_channel is None or not self._outgoing_channel.is_open or not self.connection.is_open:
                self._outgoing_channel = self.connection.channel()
                assert self._outgoing_channel is not None
                if self._publish_confirm:
                    self._outgoing_channel.confirm_delivery()

            return self._outgoing_channel
        except Exception as ex:
            raise OutputDeviceException('Could not connect to rabbitmq.') from ex

    def get_output_device(self, name: str, exchange: Optional[str] = None) -> RabbitMQOutputDevice:
        """
        Returns and outgoing device by name

        :param name: the name of the device to write to
        :param exchange: the exchange name in RabbitMQ for this output device
        :return: an output device for 'device_name'
        """
        try:
            if exchange is None:
                exchange = self._default_output_exchange

            return RabbitMQOutputDevice(self, name, exchange)
        except Exception as e:
            message = f"Couldn't create output device '{name}'"
            self._logger.exception(message)
            raise OutputDeviceException(message) from e

    def connect(self):
        """
        connects to the device manager
        """
        try:
            self._connect()
        except Exception as ex:
            raise OutputDeviceException('Could not connect to rabbitmq.') from ex

    def disconnect(self):
        """
        disconnects from the device manager
        """
        self._disconnect()
